# lsqfitgp/tests/test_kernels.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys
import inspect
import abc
import re

import pytest
import numpy as np
from scipy import linalg
from jax import test_util
import jax

sys.path = ['.'] + sys.path
import lsqfitgp as lgp
from lsqfitgp import _kernels, _Kernel

import util

# Make list of Kernel concrete subclasses.
kernels = []
for name, obj in vars(_kernels).items():
    if inspect.isclass(obj) and issubclass(obj, _Kernel.Kernel):
        assert obj not in (_Kernel.Kernel, _Kernel.StationaryKernel, _Kernel.IsotropicKernel), obj
        if name.startswith('_'):
            continue
        kernels.append(obj)

pytestmark = pytest.mark.filterwarnings(
    r'ignore:overriding init argument\(s\)',
    r'ignore:Output seems independent of input',
)

# TODO a test taking first and second derivatives w.r.t. loc and scale, which
# would apply also to non-derivable kernels. Make sure to have distances both 0
# and very near 0. Add a test option for testing a certain list of parameters,
# which defaults to loc and scale.

# TODO systematically test higher dimensions now that maxdim allows it.

class KernelTestABC(metaclass=abc.ABCMeta):
    
    @property
    @abc.abstractmethod
    def kernel_class(self):
        pass
    
    def __init_subclass__(cls):
        for name, meth in vars(cls).items():
            if name.startswith('test_'):
                setattr(cls, name, util.tryagain(meth, method=True))

class KernelTestBase(KernelTestABC):
    """
    Abstract base class to test kernels. Each subclass tests one specific
    kernel.
    """
    
    @property
    def kwargs_list(self):
        return [dict()]
    
    def random_x(self, **kw):
        return np.random.uniform(-5, 5, size=100)
    
    def random_x_nd(self, ndim, **kw):
        xs = [self.random_x(**kw) for _ in range(ndim)]
        x = np.empty(len(xs[0]), dtype=ndim * [('', xs[0].dtype)])
        for i in range(ndim):
            x[x.dtype.names[i]] = xs[i]
        return x
    
    @staticmethod
    def make_x_nd_implicit(x):
        n = len(x.dtype.names)
        dtype = x.dtype.fields[x.dtype.names[0]][0]
        newx = np.empty(x.shape, [('f0', dtype, (n,))])
        for i, name in enumerate(x.dtype.names):
            newx['f0'][..., i] = x[name]
        return newx
    
    @property
    def eps(self):
        """ used only by test_positive_* """
        return 200 * np.finfo(float).eps
        # TODO reduce this default
    
    def test_public(self):
        assert self.kernel_class in vars(lgp).values()
    
    def positive(self, deriv, nd=False):
        donesomething = False
        for kw in self.kwargs_list:
            x = self.random_x_nd(2, **kw) if nd else self.random_x(**kw)
            kernel = self.kernel_class(**kw)
            if kernel.derivable < deriv:
                continue
            if nd and kernel.maxdim < 2:
                continue
            d = (deriv, 'f0') if nd else deriv
            cov = kernel.diff(d, d)(x[None, :], x[:, None])
            util.assert_allclose(cov, cov.T, rtol=1e-5, atol=1e-7)
            eigv = linalg.eigvalsh(cov)
            assert np.min(eigv) >= -len(cov) * self.eps * np.max(eigv)
            donesomething = True
        if not donesomething:
            pytest.skip()
    
    def test_positive(self):
        self.positive(0)
    
    def test_positive_deriv(self):
        self.positive(1)
    
    def test_positive_deriv2(self):
        self.positive(2)
    
    def test_positive_nd(self):
        self.positive(0, True)
    
    def test_positive_deriv_nd(self):
        self.positive(1, True)
    
    def test_positive_deriv2_nd(self):
        self.positive(2, True)
    
    def symmetric_offdiagonal(self, xderiv, yderiv):
        donesomething = False
        for kw in self.kwargs_list:
            x = self.random_x(**kw)[None, :]
            if xderiv == yderiv:
                y = self.random_x(**kw)[:, None]
            else:
                y = x.T
            kernel = self.kernel_class(**kw)
            if kernel.derivable < max(xderiv, yderiv):
                continue
            b1 = kernel.diff(xderiv, yderiv)(x, y)
            b2 = kernel.diff(yderiv, xderiv)(y, x)
            util.assert_allclose(b1, b2, atol=1e-10)
            donesomething = True
        if not donesomething:
            pytest.skip()

    def test_symmetric_00(self):
        self.symmetric_offdiagonal(0, 0)
    
    def test_symmetric_10(self):
        self.symmetric_offdiagonal(1, 0)
    
    def test_symmetric_11(self):
        self.symmetric_offdiagonal(1, 1)
    
    def test_symmetric_20(self):
        self.symmetric_offdiagonal(2, 0)
    
    def test_symmetric_21(self):
        self.symmetric_offdiagonal(2, 1)
    
    def test_symmetric_22(self):
        self.symmetric_offdiagonal(2, 2)
    
    # TODO test higher derivatives?
    
    def jit(self, deriv=0, nd=False):
        donesomething = False
        for kw in self.kwargs_list:
            if nd:
                x = self.random_x_nd(2, **kw)
                x = lgp.StructuredArray(x)
                dtype = np.result_type(*(x[name].dtype for name in x.dtype.names))
            else:
                x = self.random_x(**kw)
                dtype = x.dtype
            if not np.issubdtype(dtype, np.number) and not dtype == bool:
                continue
            kernel = self.kernel_class(**kw)
            if kernel.derivable < deriv:
                continue
            if nd and kernel.maxdim < 2:
                continue
            d = (deriv, 'f0') if nd else deriv
            kernel = kernel.diff(d, d)
            cov1 = kernel(x[None, :], x[:, None])
            cov2 = jax.jit(kernel)(x[None, :], x[:, None])
            util.assert_allclose(cov2, cov1, rtol=1e-6, atol=1e-5)
            donesomething = True
        if not donesomething:
            pytest.skip()
    
    # TODO jit with kw as arguments. use static_argnames filtering by dtype.
    
    # TODO test vmap
    
    def test_jit(self):
        self.jit()
    
    def test_jit_deriv(self):
        self.jit(1)
    
    def test_jit_deriv2(self):
        self.jit(2)
    
    def test_jit_nd(self):
        self.jit(0, True)
    
    def test_jit_deriv_nd(self):
        self.jit(1, True)
    
    def test_jit_deriv2_nd(self):
        self.jit(2, True)

    def test_normalized(self):
        kernel = self.kernel_class
        if issubclass(kernel, _Kernel.StationaryKernel):
            for kw in self.kwargs_list:
                x = self.random_x(**kw)
                var = kernel(**kw)(x, x)
                util.assert_allclose(var, 1, rtol=1e-14, atol=1e-15)
        else:
            pytest.skip()
    
    def check_continuous_at_zero(self, deriv):
        kernel = self.kernel_class
        if not issubclass(kernel, _Kernel.StationaryKernel):
            pytest.skip()
        
        donesomething = False
        for kw in self.kwargs_list:
            t = self.random_x(**kw).dtype
            if not np.issubdtype(t, np.inexact):
                pytest.skip()
            k = kernel(**kw)
            if k.derivable < deriv:
                continue
            
            # discontinuos kernels
            if kernel == _kernels.Cauchy and kw.get('alpha', 2) < 1:
                continue
            if kernel == _kernels.GammaExp and kw.get('gamma', 1) < 1:
                continue
            if kernel == _kernels.Matern and kw['nu'] - deriv < 0.5:
                continue
            if kernel == _kernels.StationaryFracBrownian and kw.get('H', 0.5) < 0.5:
                continue
            if kernel == _kernels.White:
                continue
            if kernel == _kernels.Zeta and kw['nu'] - deriv < 0.5:
                continue
            
            k = k.diff(deriv, deriv)
            c0 = k(0, 0)
            c1 = k(0, 1e-15)
            util.assert_allclose(c1, c0, rtol=1e-10)
            
            donesomething = True
        if not donesomething:
            pytest.skip()
        
    def test_continuous_at_zero_0(self):
        self.check_continuous_at_zero(0)

    def test_continuous_at_zero_1(self):
        self.check_continuous_at_zero(1)

    def test_continuous_at_zero_2(self):
        self.check_continuous_at_zero(2)

    def test_stationary(self):
        kernel = self.kernel_class
        if issubclass(kernel, _Kernel.StationaryKernel):
            for kw in self.kwargs_list:
                x = self.random_x(**kw)
                var = kernel(**kw)(x, x)
                util.assert_allclose(var, var[0])
        else:
            pytest.skip()

    def test_double_diff_scalar_first(self):
        donesomething = False
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            if kernel.derivable < 1:
                continue
            x = self.random_x(**kw)
            r1 = kernel.diff(1, 1)(x[None, :], x[:, None])
            r2 = kernel.diff(1, 0).diff(0, 1)(x[None, :], x[:, None])
            util.assert_allclose(r1, r2)
            donesomething = True
        if not donesomething:
            pytest.skip()
    
    def test_double_diff_scalar_second(self):
        donesomething = False
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            if kernel.derivable < 2:
                continue
            x = self.random_x(**kw)
            r1 = kernel.diff(2, 2)(x[None, :], x[:, None])
            r2 = kernel.diff(1, 1).diff(1, 1)(x[None, :], x[:, None])
            util.assert_allclose(r1, r2, atol=1e-15, rtol=1e-9)
            donesomething = True
        if not donesomething:
            pytest.skip()
    
    def test_double_diff_scalar_second_chopped(self):
        donesomething = False
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            if kernel.derivable < 2:
                continue
            x = self.random_x(**kw)
            r1 = kernel.diff(2, 2)(x[None, :], x[:, None])
            r2 = kernel.diff(2, 0).diff(0, 2)(x[None, :], x[:, None])
            util.assert_allclose(r1, r2)
            donesomething = True
        if not donesomething:
            pytest.skip()
    
    def test_double_diff_nd_first(self):
        donesomething = False
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            if kernel.derivable < 1:
                continue
            if kernel.maxdim < 2:
                continue
            x = self.random_x_nd(2, **kw)[None, :]
            r1 = kernel.diff('f0', 'f0')(x, x.T)
            r2 = kernel.diff('f0', 0).diff(0, 'f0')(x, x.T)
            util.assert_allclose(r1, r2)
            donesomething = True
        if not donesomething:
            pytest.skip()

    def test_double_diff_nd_second(self):
        donesomething = False
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            if kernel.derivable < 2:
                continue
            if kernel.maxdim < 2:
                continue
            x = self.random_x_nd(2, **kw)[None, :]
            r1 = kernel.diff((2, 'f0'), (2, 'f1'))(x, x.T)
            r2 = kernel.diff('f0', 'f0').diff('f1', 'f1')(x, x.T)
            util.assert_allclose(r1, r2)
            donesomething = True
        if not donesomething:
            pytest.skip()

    def test_double_diff_nd_second_chopped(self):
        donesomething = False
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            if kernel.derivable < 2:
                continue
            if kernel.maxdim < 2:
                continue
            x = self.random_x_nd(2, **kw)[None, :]
            r1 = kernel.diff((2, 'f0'), (2, 'f1'))(x, x.T)
            r2 = kernel.diff('f0', 'f1').diff('f0', 'f1')(x, x.T)
            util.assert_allclose(r1, r2, atol=1e-15, rtol=1e-12)
            donesomething = True
        if not donesomething:
            pytest.skip()
    
    def test_implicit_fields(self):
        donesomething = False
        for kw in self.kwargs_list:
            x1 = self.random_x_nd(2, **kw)[:, None]
            x2 = self.make_x_nd_implicit(x1)
            covfun = self.kernel_class(**kw)
            if covfun.maxdim < 2:
                continue
            c1 = covfun(x1, x1.T)
            c2 = covfun(x2, x2.T)
            util.assert_allclose(c1, c2, atol=1e-15, rtol=1e-14)
            donesomething = True
        if not donesomething:
            pytest.skip()
    
    def test_loc_scale_nd(self):
        kernel = self.kernel_class
        loc = -2  # < 0
        scale = 3 # > abs(loc)
        donesomething = False
        for kw in self.kwargs_list:
            # TODO maybe put loc and scale in kw and let random_x adapt the
            # domain to loc and scale
            x = self.random_x(**kw)
            if not np.issubdtype(x.dtype, np.inexact):
                continue
            x1 = self.random_x_nd(2, **kw)[:, None]
            x2 = self.make_x_nd_implicit(x1)
            x2['f0'] -= loc
            x2['f0'] /= scale
            covfun1 = kernel(loc=loc, scale=scale, **kw)
            if covfun1.maxdim < 2:
                continue
            c1 = covfun1(x1, x1.T)
            c2 = kernel(**kw)(x2, x2.T)
            util.assert_allclose(c1, c2, rtol=1e-12, atol=1e-13)
            donesomething = True
        if not donesomething:
            pytest.skip()
    
    def test_weird_derivable(self):
        for kw in self.kwargs_list:
            kw = dict(kw)
            kw.update(derivable=[1])
            with pytest.raises(TypeError):
                self.kernel_class(**kw)
            kw.update(derivable=1.5)
            with pytest.raises(ValueError):
                self.kernel_class(**kw)
    
    def test_dim(self):
        donesomething = False
        for kw in self.kwargs_list:
            x1 = self.random_x_nd(2, **kw)[:, None]
            x2 = x1['f0']
            # dtype = x1.dtype.fields['f0'][0]
            # x3 = np.empty(x1.shape, [('f0', dtype, (1,))])
            # x3['f0'] = x2[..., None]
            x3 = self.make_x_nd_implicit(x1[['f0']])
            covfun1 = self.kernel_class(dim='f0', **kw)
            if covfun1.maxdim < 2:
                continue
            c1 = covfun1(x1, x1.T)
            c2 = self.kernel_class(**kw)(x2, x2.T)
            c3 = self.kernel_class(dim='f0', **kw)(x3, x3.T)
            util.assert_equal(c1, c2)
            util.assert_equal(c1, c3)
            with pytest.raises(ValueError):
                self.kernel_class(dim='f0', **kw)(x2, x2.T)
            donesomething = True
        if not donesomething:
            pytest.skip()
    
    def test_binary_type_error(self):
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            with pytest.raises(TypeError):
                kernel = kernel + [0]
            with pytest.raises(TypeError):
                kernel = [0] + kernel
            kernel = kernel.rescale(lambda x: 1, None) # make a _CrossKernel
            with pytest.raises(TypeError):
                kernel = kernel + [0]
            with pytest.raises(TypeError):
                kernel = [0] + kernel
    
    def test_pow(self):
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            with pytest.raises(TypeError):
                kernel = kernel ** kernel
            k1 = kernel ** 2
            k2 = kernel * kernel
            x = self.random_x(**kw)[:, None]
            c1 = k1(x, x.T)
            c2 = k2(x, x.T)
            util.assert_equal(c1, c2)
        
    def test_transf_noop(self):
        meths = ['rescale', 'xtransf', 'diff', 'fourier', 'taylor']
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            for meth in meths:
                assert kernel is getattr(kernel, meth)(None, None)
    
    def test_unknown_derivability(self):
        for kw in self.kwargs_list:
            kw = dict(kw)
            kw.update(derivable=None)
            kernel = self.kernel_class(**kw)
            assert kernel.derivable is None
    
    def test_invalid_input(self):
        kernel = self.kernel_class
        if issubclass(kernel, _Kernel.StationaryKernel):
            for kw in self.kwargs_list:
                kw = dict(kw)
                kw.update(input='cippa')
                with pytest.raises(KeyError):
                    kernel(**kw)
        else:
            pytest.skip()
    
    def test_soft_input(self):
        kernel = self.kernel_class
        donesomething = False
        if issubclass(kernel, _Kernel.StationaryKernel) and not issubclass(kernel, _Kernel.IsotropicKernel):
            for kw in self.kwargs_list:
                x1 = self.random_x(**kw)
                if not np.issubdtype(x1.dtype, np.inexact):
                    continue
                x2 = x1 - 1 / np.pi
                if np.any(np.abs(x1 - x2) < 1e-6):
                    pytest.xfail()
                kw = dict(kw)
                kw.update(input='signed')
                c1 = kernel(**kw)(x1, x2)
                kw.update(input='soft')
                c2 = kernel(**kw)(x1, x2)
                util.assert_allclose(c1, c2, atol=1e-14, rtol=1e-14)
                donesomething = True
        if not donesomething:
            pytest.skip()
    
    def test_where(self):
        donesomething = False
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            
            if kernel.maxdim < 2:
                continue
            
            x = self.random_x_nd(2, **kw)
            x0 = x['f0'][0]
            cond = lambda x: x < x0
            k1 = _Kernel.where(cond, kernel, 2 * kernel, dim='f0')
            k2 = _Kernel.where(lambda x: cond(x['f0']), kernel, 2 * kernel)
            c1 = k1(x[:, None], x[None, :])
            c2 = k2(x[:, None], x[None, :])
            
            x = self.make_x_nd_implicit(x)
            cond = lambda x: x['f0'][..., 0] < x0
            k1 = _Kernel.where(cond, kernel, 2 * kernel, dim='f0')
            k2 = _Kernel.where(cond, kernel, 2 * kernel)
            c3 = k1(x[:, None], x[None, :])
            c4 = k2(x[:, None], x[None, :])

            util.assert_equal(c1, c2)
            util.assert_equal(c3, c4)
            util.assert_equal(c1, c3)
            
            x = self.random_x(**kw)
            with pytest.raises(ValueError):
                k1(x, x)
            
            donesomething = True
        if not donesomething:
            pytest.skip()
    
    def test_rescale_swap(self):
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            x = self.random_x(**kw)[:, None]
            y = self.random_x(**kw)[None, :]
            x0 = x[0]
            f = lambda x: np.where(x < x0, -1, 1)
            k1 = kernel.rescale(f, None)
            k2 = kernel.rescale(None, f)
            c1 = k1(x, y)
            c2 = k2(y, x)
            util.assert_equal(c1, c2)
    
    def test_fourier_swap(self):
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            try:
                kernel.fourier(True, None)
            except NotImplementedError:
                pytest.skip()
            x = self.random_x(**kw)[:, None]
            k = np.arange(1, 11)[None, :]
            k1 = kernel.fourier(True, None)
            k2 = kernel.fourier(None, True)
            c1 = k1(k, x)
            c2 = k2(x, k)
            util.assert_equal(c1, c2)

    def test_xtransf_swap(self):
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            x = self.random_x(**kw)
            y = self.random_x(**kw)[None, :]
            xt = np.arange(len(x))
            f = lambda xt: x[xt]
            xt = xt[:, None]
            k1 = kernel.xtransf(f, None)
            k2 = kernel.xtransf(None, f)
            c1 = k1(xt, y)
            c2 = k2(y, xt)
            util.assert_equal(c1, c2)
    
    def test_derivable(self):
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            d = kernel.derivable
            with pytest.raises(RuntimeError):
                kernel.diff(d + 1, d + 1)
            with pytest.raises(RuntimeError):
                kernel.diff(d + 1, None)
            with pytest.raises(RuntimeError):
                kernel.diff(None, d + 1)
            if d == 0:
                continue
            d = min(d, 2)
            kernel *= _kernels.White()
            with pytest.warns(UserWarning):
              kernel.diff(1, 1)
            with pytest.warns(UserWarning):
                kernel.diff(None, d)
            with pytest.warns(UserWarning):
                kernel.diff(d, None)
    
    def test_diff_errors(self):
        donesomething = False
        for kw in self.kwargs_list:
            kw = dict(kw)
            kw.update(derivable=True)
            kernel = self.kernel_class(**kw)
            
            x = self.random_x(**kw)
            with pytest.raises(ValueError):
                kernel.diff('f0', 'f0')(x, x)
            
            x = self.random_x_nd(2, **kw)
            with pytest.raises(ValueError):
                kernel.diff('a', 'a')(x, x)
            
            dtype = x.dtype.fields['f0'][0]
            if not np.issubdtype(dtype, np.number):
                with pytest.raises(TypeError):
                    kernel.diff('f0', 'f0')(x, x)
            
            donesomething = True
        if not donesomething:
            pytest.skip()
    
    def test_fourier(self):
        for kw in self.kwargs_list:
            kernel = self.kernel_class(**kw)
            try:
                kernel.fourier(True, True)
            except NotImplementedError:
                pytest.skip()
            
            if isinstance(kernel, _kernels.Zeta) and kw['nu'] == 0:
                continue
            
            x = np.linspace(0, 1, 100)
            gp = lgp.GP(kernel, posepsfac=200)
            gp.addkernelop('fourier', True, 'F')
            gp.addx(x, 'x')
            gp.addx(1, 's1', proc='F')
            gp.addx(2, 'c1', proc='F')
            ms, cs = gp.predfromdata(dict(s1=1, c1=0), 'x', raw=True)
            mc, cc = gp.predfromdata(dict(c1=1, s1=0), 'x', raw=True)
            util.assert_allclose(ms, np.sin(2 * np.pi * x), atol=1e-15)
            util.assert_allclose(mc, np.cos(2 * np.pi * x), atol=1e-15)
            util.assert_allclose(np.diag(cs), cs[0, 0], atol=1e-15)
            util.assert_allclose(np.diag(cc), cc[0, 0], atol=1e-15)
    
    @classmethod
    def make_subclass(cls, kernel_class, kwargs_list=None, random_x_fun=None, eps=None):
        name = 'Test' + kernel_class.__name__
        subclass = type(cls)(name, (cls,), {
            'kernel_class': property(lambda self: kernel_class)
        })
        if kwargs_list is not None:
            subclass.kwargs_list = property(lambda self: kwargs_list)
        if random_x_fun is not None:
            subclass.random_x = lambda self, **kw: random_x_fun(**kw)
        if eps is not None:
            subclass.eps = eps
        return subclass

def matrix_square(A):
    return A.T @ A

def random_nd(size, ndim):
    out = np.empty(size, dtype=[('xyz', float, (ndim,))])
    out['xyz'] = np.random.uniform(-5, 5, size=(size, ndim))
    return out

lipsum = """

Duis mollis, est non commodo luctus, nisi erat porttitor ligula, eget lacinia
odio sem nec elit. Curabitur blandit tempus porttitor. Sed posuere consectetur
est at lobortis. Cras mattis consectetur purus sit amet fermentum.

Morbi leo risus, porta ac consectetur ac, vestibulum at eros. Cras mattis
consectetur purus sit amet fermentum. Lorem ipsum dolor sit amet, consectetur
adipiscing elit. Donec ullamcorper nulla non metus auctor fringilla.

Cras justo odio, dapibus ac facilisis in, egestas eget quam. Donec ullamcorper
nulla non metus auctor fringilla. Praesent commodo cursus magna, vel
scelerisque nisl consectetur et. Donec id elit non mi porta gravida at eget
metus. Nullam id dolor id nibh ultricies vehicula ut id elit. Maecenas sed diam
eget risus varius blandit sit amet non magna.

Morbi leo risus, porta ac consectetur ac, vestibulum at eros. Sed posuere
consectetur est at lobortis. Donec ullamcorper nulla non metus auctor
fringilla. Fusce dapibus, tellus ac cursus commodo, tortor mauris condimentum
nibh, ut fermentum massa justo sit amet risus. Lorem ipsum dolor sit amet,
consectetur adipiscing elit. Vivamus sagittis lacus vel augue laoreet rutrum
faucibus dolor auctor.

"""

lipsum_words = np.array(list(set(re.split(r'\s|[,.]', lipsum.lower()))))

def bow_rand(**kw):
    return np.array([' '.join(np.random.choice(lipsum_words, 10)) for _ in range(30)])
        
# Define a concrete subclass of KernelTestBase for each kernel.
test_kwargs = {
    _kernels.Matern: dict(kwargs_list=
        [dict(nu=v + 0.5 ) for v in range(   5)] +
        [dict(nu=v + 0.49) for v in range(   5)] +
        [dict(nu=v + 0.51) for v in range(   5)] +
        [dict(nu=v       ) for v in range(   5)] +
        [dict(nu=v - 0.01) for v in range(1, 5)] +
        [dict(nu=v + 0.01) for v in range(   5)]
    ),
    _kernels.Maternp: dict(kwargs_list=[dict(p=p) for p in range(10)]),
    _kernels.Wendland: dict(kwargs_list=[
        dict(k=k, alpha=a) for k in range(4) for a in np.linspace(1, 4, 10)
    ], eps=1e3 * np.finfo(float).eps),
    _kernels.Wiener: dict(random_x_fun=lambda **kw: np.random.uniform(0, 10, size=100)),
    _kernels.WienerIntegral: dict(random_x_fun=lambda **kw: np.random.uniform(0, 10, size=100)),
    _kernels.FracBrownian: dict(
        random_x_fun=lambda **kw: np.random.uniform(-10, 10, size=100),
        kwargs_list=[dict(H=H, K=K) for H in [0.1, 0.5, 1] for K in [0.1, 0.5, 1]],
    ),
    _kernels.BrownianBridge: dict(random_x_fun=lambda **kw: np.random.uniform(0, 1, size=100)),
    _kernels.OrnsteinUhlenbeck: dict(random_x_fun=lambda **kw: np.random.uniform(0, 10, size=100)),
    _kernels.Categorical: dict(kwargs_list=[
        dict(cov=matrix_square(np.random.randn(10, 10)))
    ], random_x_fun=lambda **kw: np.random.randint(10, size=100)),
    _kernels.NNKernel: dict(eps=4 * np.finfo(float).eps),
    _kernels.Zeta: dict(kwargs_list=[
        dict(nu=v) for v in [0, 0.1, 1, 1.5, 4.9, 1000]
    ]),
    _kernels.Celerite: dict(kwargs_list=[
        dict(), dict(gamma=1, B=1), dict(gamma=0, B=0), dict(gamma=10, B=0)
    ]),
    _kernels.Harmonic: dict(kwargs_list=[
        dict(), dict(Q=0.01), dict(Q=0.25), dict(Q=0.75), dict(Q=0.99), dict(Q=1), dict(Q=1.01), dict(Q=2)
    ]),
    _kernels.BagOfWords: dict(random_x_fun=bow_rand),
    _kernels.Gibbs: dict(kwargs_list=[dict(derivable=True)]),
    _kernels.Rescaling: dict(kwargs_list=[dict(derivable=True)]),
    _kernels.GammaExp: dict(kwargs_list=[dict(), dict(gamma=2)], eps=1e3 * np.finfo(float).eps),
    _kernels.Bessel: dict(kwargs_list=[dict()] + [dict(nu=nu) for nu in range(5)] + [dict(nu=nu - 0.01) for nu in range(1, 5)] + [dict(nu=nu + 0.01) for nu in range(5)] + [dict(nu=nu + 0.5) for nu in range(5)]),
    _kernels.Cauchy: dict(kwargs_list=[dict(alpha=a, beta=b) for a in [0.001, 0.5, 0.999, 1, 1.001, 1.5, 1.999, 2] for b in [0.001, 0.5, 1, 1.5, 2, 4, 8]]),
    _kernels.CausalExpQuad: dict(kwargs_list=[dict(alpha=a) for a in [0, 1, 2]]),
    _kernels.Decaying: dict(
        random_x_fun=lambda **_: np.random.uniform(0, 5, size=100),
        kwargs_list=[dict(alpha=a) for a in [0, .5, 1, 2]],
    ),
    _kernels.StationaryFracBrownian: dict(kwargs_list=[dict(H=H) for H in [0.1, 0.5, 1]]),
    _kernels.Circular: dict(kwargs_list=[dict(c=c, tau=t) for c, t in [(0.1, 4), (0.5, 4), (0.5, 8)]]),
    _kernels.MA: dict(
        random_x_fun=lambda **_: np.random.randint(0, 100, 100),
        kwargs_list=[dict(w=w) for w in [
            [], [0], [1], [1, 1], [1, -1], [2, 1], [1, 2, 3, 4, 5], np.random.randn(30),
        ]],
    ),
    _kernels.AR: dict(
        random_x_fun=lambda **_: np.random.randint(0, 100, 100),
        kwargs_list=[dict(phi=phi, maxlag=100) for phi in [
            [], [0], [-0.5], [0.5], [0.9], [-0.9], [0.5, 0], [0, 0.5], 3 * [0] + [0.5],
        ]] + [dict(gamma=gamma, maxlag=100) for gamma in [
            [0], [1], [1, 0], [1, 0.5], [1, 0.5, 0.25], [1, -0.9],
        ]] + [dict(slnr=r, lnc=c) for r, c in [
            ([], []),
            ([1/10], []),
            ([1/2], []),
            ([1/10, 1/2], []),
            ([1/10, 1/10], []),
            ([1/10, 1/10, -1/2], []),
            ([], [1/10 + 1j]),
            ([], [1/2 + 1j]),
            ([], [1/10 + 1j, 1/2 + 2j]),
            ([], [1/10 + 1j, 1/10 + 2j]),
            ([], [1/10 + 1j, 1/10 + 1j, 1/2 + 2j]),
            ([1/10, 1/10, -1/2], [1/10 + 1j, 1/10 + 1j, 1/2 + 2j]),
        ]],
    ),
    _kernels.Color: dict(kwargs_list=[dict(n=n) for n in [2, 3, 4, 5, 6, 20]]),
    _kernels.BART: dict(kwargs_list=[
        dict(
            splits=_kernels.BART.splits_from_coord(np.random.randn(10, 1)),
            alpha=a,
            beta=b,
            maxd=d,
            reset=r,
        )
        for a in [0., 1., 0.95]
        for b in [0, 1, 2, 10]
        for d in [0, 1, 2, 3]
        for r in [None, (d + 1) // 2]
    ]),
    # TODO I need a way to use nd splits only with nd x
}

for kernel in kernels:
    factory_kw = test_kwargs.get(kernel, {})
    newclass = KernelTestBase.make_subclass(kernel, **factory_kw)
    exec('{} = newclass'.format(newclass.__name__))

def check_matern_half_integer(deriv):
    """
    Check that the formula for half integer nu gives the same result of the
    formula for real nu.
    """
    for p in range(10):
        x = 3 * np.random.randn(100)[None, :]
        y = x.T
        k = _kernels.Matern(nu=p + 1/2)
        d = min(k.derivable, deriv)
        r1 = k.diff(d, d)(x, y)
        r2 = _kernels.Maternp(p=p).diff(d, d)(x, y)
        util.assert_allclose(r1, r2, rtol=1e-9, atol=1e-16)

def test_matern_half_integer_0():
    check_matern_half_integer(0)

def test_matern_half_integer_1():
    check_matern_half_integer(1)

def test_matern_half_integer_2():
    check_matern_half_integer(2)
    
def test_wiener_integral():
    """
    Test that the derivative of the Wiener integral is the Wiener.
    """
    x, y = np.abs(np.random.randn(2, 100))
    r1 = _kernels.Wiener()(x, y)
    r2 = _kernels.WienerIntegral().diff(1, 1)(x, y)
    util.assert_allclose(r1, r2)

def test_celerite_harmonic():
    """
    Check that the Celerite kernel is equivalent to the Harmonic kernel when
    B == gamma.
    """
    x = np.random.uniform(-1, 1, size=100)
    Q = np.random.uniform(1.1, 3)
    eta = np.sqrt(1 - 1 / Q**2)
    B = 1 / (eta * Q)
    r1 = _kernels.Celerite(gamma=B, B=B)(x[:, None], x[None, :])
    r2 = _kernels.Harmonic(Q=Q, scale=eta)(x[:, None], x[None, :])
    util.assert_allclose(r1, r2, atol=1e-15, rtol=1e-15)

def check_harmonic_continuous(deriv, Q0, Qderiv=False):
    eps = 1e-10
    Q0 = float(Q0)
    Qs = [(1 - eps) * Q0, Q0, (1 + eps) * Q0]
    x = np.random.randn(100)
    results = []
    for Q in Qs:
        kernelf = lambda Q, x: _kernels.Harmonic(Q=Q).diff(deriv, deriv)(x[None, :], x[:, None])
        if Qderiv:
            kernelf = jax.jacfwd(kernelf)
        results.append(kernelf(Q, x))
    util.assert_allclose(results[0], results[2], atol=1e-5)
    util.assert_allclose(results[0], results[1], atol=1e-5)
    util.assert_allclose(results[1], results[2], atol=1e-5)

def test_harmonic_continuous_12():
    check_harmonic_continuous(0, 1/2)
def test_harmonic_continuous_1():
    check_harmonic_continuous(0, 1)
def test_harmonic_deriv_continuous_12():
    check_harmonic_continuous(1, 1/2)
def test_harmonic_deriv_continuous_1():
    check_harmonic_continuous(1, 1)
def test_harmonic_derivQ_continuous_12():
    check_harmonic_continuous(0, 1/2, True)
def test_harmonic_derivQ_continuous_1():
    check_harmonic_continuous(0, 1, True)
def test_harmonic_deriv_derivQ_continuous_12():
    check_harmonic_continuous(1, 1/2, True)
def test_harmonic_deriv_derivQ_continuous_1():
    check_harmonic_continuous(1, 1, True)

def test_nonfloat_eps():
    x = np.arange(20)
    c1 = _kernels.Wendland()(x, x)
    eps = np.finfo(float).eps
    c2 = np.exp(-eps)
    util.assert_allclose(c1, c2, rtol=eps, atol=eps)

def test_default_override():
    with pytest.warns(UserWarning):
        _kernels.Maternp(p=0, derivable=17)

def test_kernel_decorator_error():
    with pytest.raises(ValueError):
        _Kernel.kernel(1, 2)

def test_transf_not_implemented():
    meths = ['fourier', 'taylor']
    kernel = _kernels.Maternp(p=0)
    for meth in meths:
        with pytest.raises(NotImplementedError):
            getattr(kernel, meth)(True, None)
        with pytest.raises(NotImplementedError):
            getattr(kernel, meth)(None, True)
        with pytest.raises(NotImplementedError):
            getattr(kernel, meth)(True, True)

def test_maxdim():
    _Kernel.Kernel(lambda x, y: 1, maxdim=None)
    kernel = _Kernel.Kernel(lambda x, y: 1, maxdim=0)
    with pytest.raises(ValueError):
        kernel(0, 0)
    kernel = _Kernel.Kernel(lambda x, y: 1, maxdim=1)
    with pytest.raises(ValueError):
        z = np.zeros((), 'd,d')
        kernel(z, z)

def test_nary():
    kernel = _Kernel.Kernel(lambda x, y: np.maximum(0, 1 - np.abs(x - y)))
    g = lambda x: 2 * x
    k1 = _Kernel.Kernel._nary(lambda f: lambda x: g(x) * f(x), [kernel], kernel._side.LEFT)
    k2 = _Kernel.Kernel._nary(lambda f: lambda y: g(y) * f(y), [kernel], kernel._side.RIGHT)
    x = np.linspace(-5, 5, 30)
    m1 = k1(x[:, None], x[None, :])
    assert not np.allclose(m1, m1.T)
    m2 = k2(x[:, None], x[None, :])
    assert not np.allclose(m2, m2.T)
    assert not np.allclose(m1, m2)
    util.assert_close_matrices(m1, m2.T)

def test_wendland_highk():
    kernel = _kernels.Wendland(k=4)
    with pytest.raises(NotImplementedError):
        kernel(0, 0)

def test_default_transf():
    k1 = _Kernel.Kernel(lambda x, y: x * y, loc=1, maxdim=None)
    k2 = _Kernel.Kernel(lambda x, y: x * y, loc=1)
    x = np.linspace(-5, 3, 10)[:, None]
    v1 = k1(x, x.T)
    v2 = k2(x, x.T)
    util.assert_equal(v1, v2)

#####################  XFAILS/SKIPS  #####################

util.skip(TestAR, 'test_normalized')
util.skip(TestMA, 'test_normalized')

# TODO These are isotropic kernels with the input='soft' option. The problems
# arise where x == y. => use make_jaxpr to debug?
util.xfail(TestWendland, 'test_positive_deriv2_nd')
util.xfail(TestWendland, 'test_double_diff_nd_second_chopped')
util.xfail(TestWendland, 'test_continuous_at_zero_2')
util.xfail(TestWendland, 'test_jit_deriv2_nd') # seen xpassing, precision?
util.xfail(TestCausalExpQuad, 'test_positive_deriv2_nd')
util.xfail(TestCausalExpQuad, 'test_double_diff_nd_second_chopped')
util.xfail(TestCausalExpQuad, 'test_continuous_at_zero_2')

# TODO some xpass, likely numerical precision problems
util.xfail(TestWendland, 'test_positive_deriv2') # normally xpasses
util.xfail(TestCausalExpQuad, 'test_positive_deriv2') # NOT 1 - erf cancel

# TODO This one should not fail, it's a first derivative! Probably it's the
# case D = 1 that fails because that's the maximum dimensionality. For some
# reason I don't catch it without taking a derivative. => This explanation is
# likely wrong since the jit test fails too, without checking positivity.
util.xfail(TestWendland, 'test_positive_deriv_nd') # seen xpassing in the wild
util.xfail(TestWendland, 'test_jit_deriv_nd') # seen xpassing, precision?

# TODO These are not isotropic kernels, what is the problem?
util.xfail(TestTaylor, 'test_double_diff_nd_second') # numerical precision
util.xfail(TestNNKernel, 'test_double_diff_nd_second') # large difference
util.xfail(TestFracBrownian, 'test_double_diff_nd_second') # large difference



from jax.interpreters import ad
from jax._src.ad_util import stop_gradient_p
import jax

@jax.custom_jvp
def f(x):
    return x ** 2

@f.defjvp
def f_jvp(x, t):
    return x[0], jax.lax.stop_gradient(t[0])

ad.primitive_transposes[stop_gradient_p] = lambda ct, _: [jax.lax.stop_gradient(ct)]

print(jax.grad(f)(1.)) # -> 1
print(jax.grad(jax.grad(f))(1.)) # -> 0
print(jax.jacfwd(jax.jacrev(f))(1.))



# lsqfitgp/examples/fourier.py
#
# Copyright (c) 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np
import gvar

"""Constrain the values of Fourier series coefficients"""

gp = lgp.GP(lgp.Zeta(nu=2.5), checkpos=False) # TODO is this checkpos necessary
gp.addkernelop('fourier', True, 'F')
x = np.linspace(0, 1, 100)
gp.addx(x, 'x')
gp.addx(1, 's1', proc='F')
gp.addx(2, 'c1', proc='F')

comb = [
    [0, 0],
    [1, 0],
    [0, 1],
    [1, 1],
]

fig, ax = plt.subplots(num='fourier', clear=True)

for s, c in comb:
    y = gp.predfromdata(dict(s1=s, c1=c), 'x')
    m = gvar.mean(y)
    u = gvar.sdev(y)
    pc = ax.fill_between(x, m - u, m + u, alpha=0.5, label=f's{s}c{c}')
    color = pc.get_facecolor()
    for sample in gvar.raniter(y, 3):
        ax.plot(x, sample, color=color)

ax.legend()

fig.show()



import jax
from jax import numpy as jnp
from jax.interpreters import ad
from jax._src.ad_util import stop_gradient_p

# ad.primitive_transposes[stop_gradient_p] = lambda ct, _: [jax.lax.stop_gradient(ct)]

@jax.custom_jvp
def stop_hessian(x):
    return x

@stop_hessian.defjvp
def stop_hessian_jvp(primals, tangents):
    x, = primals
    x_dot, = tangents
    return x, jax.lax.stop_gradient(x_dot)

def f(x):
    return jnp.sin(jnp.cos(x))

def g(x):
    return 1/2 * x ** 2

def h(x):
    return g(stop_hessian(f(x)))

# h = 1/2 f^2
# h' = f f'
# h'' = f' f' + f f'' =
#               ^^^^^ stop_hessian should remove this

x = jnp.arange(4.)
f1 = jax.vmap(jax.grad(f))
f2 = jax.vmap(jax.grad(jax.grad(f)))
print("f' f':  ", f1(x) ** 2)
print("f f'':  ", f(x) * f2(x))
print('total:  ', f1(x) ** 2 + f(x) * f2(x))
print('fwd-fwd:', jax.vmap(jax.jacfwd(jax.jacfwd(h)))(x)) # ok -> f' f'
print('rev-fwd:', jax.vmap(jax.jacrev(jax.jacfwd(h)))(x)) # ok -> f' f'
print('fwd-rev:', jax.vmap(jax.jacfwd(jax.jacrev(h)))(x)) # wrong -> f f'' (?)
print('rev-rev:', jax.vmap(jax.jacrev(jax.jacrev(h)))(x)) # wrong -> f f'' (?)



# lsqfitgp/tests/test_GP.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import copy
import sys
import itertools

import pytest
from pytest import mark
import numpy as np
from jax import numpy as jnp
import gvar
import jax

sys.path = ['.'] + sys.path
import lsqfitgp as lgp
import util

gen = np.random.default_rng(202208251700)

def test_prior_raw_shape():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(np.arange(20).reshape(2, 10), 'x')
    
    cov = gp.prior(raw=True)
    assert cov['x', 'x'].shape == (2, 10, 2, 10)
    
    cov = gp.prior('x', raw=True)
    assert cov.shape == (2, 10, 2, 10)

@mark.parametrize('shape', [(20,), (2, 3)])
def test_halfmatrix(shape):
    covs = []
    x = gen.standard_normal(shape)
    for checksym in [False, True]:
        gp = lgp.GP(lgp.ExpQuad(), checksym=checksym, halfmatrix=not checksym)
        gp.addx({'x': x})
        covs.append(gp.prior('x', raw=True))
    util.assert_equal(*covs)

def test_transf_scalar():
    gp = lgp.GP(lgp.ExpQuad() + lgp.Cauchy())
    x = np.arange(20)
    gp.addx(x, 'x')
    cov1 = gp.prior('x', raw=True)
    
    gp = lgp.GP(lgp.where(lambda f: f, lgp.ExpQuad(dim='f1'), lgp.Cauchy(dim='f1'), dim='f0'))
    y = np.empty((2, len(x)), '?,f8')
    y['f0'] = np.reshape([0, 1], (2, 1))
    y['f1'] = x
    gp.addx(y[0], 'y0')
    gp.addx(y[1], 'y1')
    gp.addtransf({'y0': 1, 'y1': 1}, 'x')
    cov2 = gp.prior('x', raw=True)
    
    util.assert_equal(cov1, cov2)

def test_transf_vector():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx([0, 1], 'x')
    gp.addtransf({'x': [-1, 1]}, 'y')
    prior = gp.prior()
    y1 = prior['y']
    y2 = prior['x'][1] - prior['x'][0]
    util.assert_same_gvars(y1, y2, atol=1e-12)

def test_cov():
    A = gen.standard_normal((20, 20))
    M1 = A.T @ A
    gp = lgp.GP()
    gp.addcov(M1, 'M')
    M2 = gp.prior('M', raw=True)
    util.assert_equal(M1, M2)

def test_compare_transfs():
    x = np.arange(20)
    def preparegp():
        gp = lgp.GP()
        gp.addproc(lgp.ExpQuad(), 'a')
        gp.addproc(lgp.Cauchy(), 'b')
        gp.addx(x, 'ax', proc='a')
        gp.addx(x, 'bx', proc='b')
        return gp
    def finalizegp(gp):
        gp.addx(x, 2, proc='ab1')
        gp.addx(x, 3, proc='ab2')
    def checkgp(gp):
        keys = [0, 1, 2, 3]
        prior = gp.prior(keys, raw=True)
        for k1 in keys:
            for k2 in keys:
                util.assert_allclose(prior[0, 0], prior[k1, k2], atol=1e-15, rtol=1e-15)
    
    # with functions
    gp = preparegp()
    fa = lambda x: jnp.sin(x) + 0.5 * jnp.cos(x ** 2)
    fb = lambda x: 1 / (1 + jnp.exp(-x))
    gp.addtransf({'ax': np.diag(fa(x)), 'bx': np.diag(fb(x))}, 0)
    gp.addlintransf(lambda a, b: fa(x) * a + fb(x) * b, ['ax', 'bx'], 1)
    gp.addproctransf({'a': fa, 'b': fb}, 'ab1')
    gp.addproclintransf(lambda a, b: lambda x: fa(x) * a(x) + fb(x) * b(x), ['a', 'b'], 'ab2')
    finalizegp(gp)
    checkgp(gp)
    
    # with scalars
    gp = preparegp()
    gp.addtransf({'ax': 2, 'bx': 3}, 0)
    gp.addlintransf(lambda a, b: 2 * a + 3 * b, ['ax', 'bx'], 1)
    gp.addproctransf({'a': 2, 'b': 3}, 'ab1')
    gp.addproclintransf(lambda a, b: lambda x: 2 * a(x) + 3 * b(x), ['a', 'b'], 'ab2')
    finalizegp(gp)
    checkgp(gp)

def test_lintransf_checks():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(0, 0)
    gp.addx(0, 1)
    with pytest.raises(KeyError):
        gp.addlintransf(lambda x, y: x + y, [0, 1], 0)
    with pytest.raises(ValueError):
        gp.addlintransf(lambda x, y: x + y, [0, 1], None)
    with pytest.raises(KeyError):
        gp.addlintransf(lambda x, y: x + y, [0, 2], 2)
    with pytest.raises(RuntimeError):
        gp.addlintransf(lambda x, y: 1 + x + y, [0, 1], 2)
    with pytest.raises(RuntimeError):
        gp.addlintransf(lambda x, y: 1, [0, 1], 2)
    gp.addlintransf(lambda x, y: 1 + x + y, [0, 1], 2, checklin=False)
    gp._checklin = False
    gp.addlintransf(lambda x, y: 1 + x + y, [0, 1], 3)
    with pytest.raises(RuntimeError):
        gp.addlintransf(lambda x, y: 1 + x + y, [0, 1], 4, checklin=True)

def test_proclintransf_checks():
    def makegp(**kw):
        gp = lgp.GP(**kw)
        gp.addproc(lgp.ExpQuad(), 0)
        gp.addproc(lgp.ExpQuad(), 1)
        return gp
    gp = makegp()
    with pytest.raises(KeyError):
        gp.addproclintransf(lambda f, g: lambda x: f(x) + g(x), [0, 1], 0)
    with pytest.raises(KeyError):
        gp.addproclintransf(lambda f, g: lambda x: f(x) + g(x), [0, 2], 2)
    with pytest.raises(RuntimeError):
        gp.addproclintransf(lambda f, g: lambda x: 1, [0, 1], 2, checklin=True)
    with pytest.raises(RuntimeError):
        gp.addproclintransf(lambda f, g: lambda x: 1 + f(x) + g(x), [0, 1], 2, checklin=True)
    with pytest.raises(RuntimeError):
        gp.addproclintransf(lambda f, g: lambda x: f(x) + g(x)[None, :], [0, 1], 2, checklin=True)
    gp.addproclintransf(lambda f, g: lambda x: 1 + f(x) + g(x), [0, 1], 2)
    gp.addproclintransf(lambda f, g: lambda x: f(x) + g(x), [0, 1], 3, checklin=True)
    gp = makegp(checklin=True)
    with pytest.raises(RuntimeError):
        gp.addproclintransf(lambda f, g: lambda x: 1, [0, 1], 2, checklin=None)

def test_proclintransf_mockup():
    def makegp(**kw):
        gp = lgp.GP(**kw)
        gp.addproc(lgp.ExpQuad(), 0)
        gp.addproc(lgp.ExpQuad(), 1)
        return gp
    gp = makegp()
    with pytest.raises(RuntimeError):
        gp.addproclintransf(lambda f, g: lambda x: 1 + f(x['dim1']) + g(x['dim2']), [0, 1], 2, checklin=True)
    gp.addproclintransf(lambda f, g: lambda x: f(x['dim1']) + g(x['dim2']), [0, 1], 2, checklin=True)

def test_lintransf_matmul():
    gp = lgp.GP(lgp.ExpQuad())
    x = np.arange(20)
    gp.addx(x, 0)
    m = gen.standard_normal((30, len(x)))
    gp.addlintransf(lambda x: m @ x, [0], 1)
    prior = gp.prior([0, 1], raw=True)
    util.assert_allclose(m @ prior[0, 0] @ m.T, prior[1, 1], rtol=1e-12)
    util.assert_allclose(m @ prior[0, 1], prior[1, 1], rtol=1e-12)
    util.assert_allclose(prior[1, 0] @ m.T, prior[1, 1], rtol=1e-12)

def test_prior_gvar():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(gen.standard_normal(20), 0)
    gp.addx(gen.standard_normal(15), 1)
    gp.addtransf({0: gen.standard_normal((15, 20))}, 2)
    gp.addlintransf(lambda x, y: jnp.real(jnp.fft.rfft(x + y)), [1, 2], 3)
    m = gen.standard_normal((40, 40))
    gp.addcov(m @ m.T, 4)
    gp.addtransf({4: gen.standard_normal((8, 40)), 3: np.pi}, 5)
    covs = gp.prior(raw=True)
    prior = gp.prior()
    gmeans = gvar.mean(prior)
    for g in gmeans.values():
        np.testing.assert_equal(g, np.zeros_like(g))
    gcovs = gvar.evalcov(prior)
    for k, cov in covs.items():
        gcov = gcovs[k]
        util.assert_close_matrices(cov, gcov, atol=1e-15, rtol=1e-9)

def test_kernelop():
    gp = lgp.GP()
    gp.addproc(lgp.ExpQuad(), 'a')
    f = lambda x: x
    gp.addproc(lgp.ExpQuad() * lgp.Rescaling(stdfun=f), 'b1')
    gp.addkernelop('rescale', f, 'b2', 'a')
    x = np.arange(20)
    gp.addx(x, 'x1', proc='b1')
    gp.addx(x, 'x2', proc='b2')
    prior = gp.prior(['x1', 'x2'], raw=True)
    util.assert_equal(prior['x1', 'x1'], prior['x2', 'x2'])
    util.assert_equal(prior['x2', 'x1'], prior['x1', 'x2'].T)
    util.assert_equal(prior['x1', 'x2'], np.zeros(2 * x.shape))

def test_not_kernel():
    with pytest.raises(TypeError):
        gp = lgp.GP(0)

def test_two_procs():
    gp = lgp.GP()
    gp.addproc(lgp.ExpQuad(), 'a')
    with pytest.raises(KeyError):
        gp.addproc(lgp.ExpQuad(), 'a')

def test_default_kernel():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addproc(key='a')
    x = np.arange(20)
    gp.addx(x, 'x1')
    gp.addx(x, 'x2', proc='a')
    prior = gp.prior(raw=True)
    util.assert_equal(prior['x1', 'x1'], prior['x2', 'x2'])
    util.assert_equal(prior['x1', 'x2'], np.zeros(2 * x.shape))

def test_no_proc():
    gp = lgp.GP()
    with pytest.raises(KeyError):
        gp.addproctransf({'a': 1}, 'b')

def test_invalid_factor():
    gp = lgp.GP()
    gp.addproc(lgp.ExpQuad(), 'a')
    with pytest.raises(TypeError):
        gp.addproctransf({'a': None}, 'b')

def test_existing_proc():
    gp = lgp.GP()
    gp.addproc(lgp.ExpQuad(), 'a')
    with pytest.raises(KeyError):
        gp.addproctransf({'a': 1}, 'a')

def test_empty_proc():
    gp = lgp.GP()
    gp.addproctransf({}, 'a')
    gp.addproclintransf(lambda: lambda x: 0, [], 'b')
    x = np.arange(20)
    gp.addx(x, 'ax', proc='a')
    gp.addx(x, 'bx', proc='b')
    cov = gp.prior(raw=True)
    util.assert_equal(cov['ax', 'ax'], np.zeros(2 * x.shape))
    util.assert_equal(cov['bx', 'bx'], np.zeros(2 * x.shape))

def test_no_op():
    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(ValueError):
        gp.addkernelop('cippa', None, 'a')

def test_already_defined():
    gp = lgp.GP()
    gp.addproc(lgp.ExpQuad(), 'a')
    with pytest.raises(KeyError):
        gp.addkernelop('diff', 1, 'a', 'a')

def test_proc_not_found():
    gp = lgp.GP()
    with pytest.raises(KeyError):
        gp.addkernelop('diff', 1, 'a', 'b')

def test_addprocderiv():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addprocderiv(1, 'a')
    x = np.arange(20)
    gp.addx(x, 0, proc='a')
    gp.addx(x, 1, deriv=1)
    prior = gp.prior(raw=True)
    util.assert_equal(prior[0, 0], prior[1, 1])
    util.assert_equal(prior[0, 0], prior[0, 1])

def test_addprocxtransf():
    gp = lgp.GP()
    f = lambda x: x ** 2
    gp.addproc(lgp.ExpQuad(), 0)
    gp.addprocxtransf(f, 'a', 0)
    gp.addproclintransf(lambda g: lambda x: g(f(x)), [0], 'b')
    x = np.linspace(0, 4, 20)
    gp.addx(x, 0, proc='a')
    gp.addx(x, 1, proc='b')
    gp.addx(f(x), 2, proc=0)
    prior = gp.prior(raw=True)
    for i in range(3):
        for j in range(3):
            util.assert_equal(prior[0, 0], prior[i, j])

def test_addprocrescale():
    gp = lgp.GP()
    s = lambda x: x ** 2
    gp.addproc(lgp.ExpQuad(), 0)
    gp.addprocrescale(s, 'a', 0)
    gp.addproclintransf(lambda f: lambda x: s(x) * f(x), [0], 'b')
    x = np.linspace(0, 2, 20)
    gp.addx(x, 'base', proc=0)
    gp.addtransf({'base': s(x) * np.eye(len(x))}, 0)
    gp.addx(x, 1, proc='a')
    gp.addx(x, 2, proc='b')
    prior = gp.prior(raw=True)
    for i in range(3):
        for j in range(3):
            util.assert_allclose(prior[0, 0], prior[i, j], rtol=1e-15, atol=1e-15)

def test_missing_proc():
    gp = lgp.GP()
    with pytest.raises(KeyError):
        gp.addx(0, 0, proc='cippa')

def test_no_key():
    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(ValueError):
        gp.addx(0)
    with pytest.raises(ValueError):
        gp.addcov(1)

def test_redundant_key():
    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(ValueError):
        gp.addx({0: 0}, 0)
    with pytest.raises(ValueError):
        gp.addcov({0: 1}, 0)

def test_none_key():
    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(ValueError):
        gp.addx({None: 0})
    
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(0, 0)
    with pytest.raises(ValueError):
        gp.addtransf({0: 1}, None)

    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(ValueError):
        gp.addcov({None: 1})

def test_nonsense_x():
    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(TypeError):
        gp.addcov(None, 0)

def test_key_already_used():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(0, 0)
    with pytest.raises(KeyError):
        gp.addx(0, 0)

    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(0, 0)
    with pytest.raises(KeyError):
        gp.addtransf({0: 1}, 0)

    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(0, 0)
    with pytest.raises(KeyError):
        gp.addcov(1, 0)

# TODO this fails in numpy 1.20, may work as versions are bumped
# def test_bad_array():
#     gp = lgp.GP(lgp.ExpQuad())
#     with pytest.raises(ValueError):
#         gp.addx({0: [[1, 2], 3]})

def test_not_empty():
    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(ValueError):
        gp.addx([], 0)

def test_incompatible_dtypes():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(0, 0)
    with pytest.raises(TypeError):
        gp.addx(np.zeros(1, 'd,d'), 1)
    
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(np.zeros(1, 'd,d'), 0)
    # gp.addx(np.zeros(1, 'i,i'), 1) # succeeds only if numpy >= 1.23
    with pytest.raises(TypeError):
        gp.addx(np.zeros(1, 'd,d,d'), 2)

def test_explicit_deriv():
    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(ValueError):
        gp.addx(0, 0, deriv='x')

def test_missing_field():
    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(ValueError):
        gp.addx(np.array((0, 0), 'f8,f8'), 0, deriv='x')

def test_missing_key():
    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(KeyError):
        gp.addtransf({0: 1}, 1)

def test_nonsense_tensors():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(0, 0)
    with pytest.raises(TypeError):
        gp.addtransf({0: 'a'}, 1)
    with pytest.raises(ValueError):
        gp.addtransf({0: np.inf}, 1)
    gp.addx([0, 1], 1)
    with pytest.raises(ValueError):
        gp.addtransf({1: [1, 2, 3]}, 2)

def test_fail_broadcast():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx([0, 1], 0)
    gp.addx([0, 1, 2], 1)
    with pytest.raises(ValueError):
        gp.addtransf({0: 1, 1: 1}, 2)

def test_addcov_wrong_blocks():
    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(ValueError):
        gp.addcov(np.zeros((1, 1, 1)), 0)
    with pytest.raises(ValueError):
        gp.addcov(np.zeros((1, 2, 2, 1)), 0)
    with pytest.raises(ValueError):
        gp.addcov(gen.standard_normal((10, 10)), 0)
    with pytest.raises(KeyError):
        gp.addcov({
            (0, 0): 1,
            (0, 1): 0,
        })
    with pytest.raises(ValueError):
        gp.addcov({
            (0, 0): np.ones((2, 2)),
            (1, 1): np.ones((3, 3)),
            (0, 1): np.ones((3, 2)),
        })
    with pytest.raises(ValueError):
        gp.addcov({
            (0, 0): np.ones((2, 2)),
            (1, 1): np.ones((3, 3)),
            (0, 1): np.ones((2, 3)),
            (1, 0): np.zeros((3, 2)),
        })

def test_addcov_no_checksym():
    gp = lgp.GP(lgp.ExpQuad(), checksym=False)
    gp.addcov({
        (0, 0): np.ones((2, 2)),
        (1, 1): np.ones((3, 3)),
        (0, 1): np.ones((2, 3)),
        (1, 0): np.zeros((3, 2)),
    })

def test_addcov_missing_block():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addcov({
        (0, 0): np.ones((2, 2)),
        (1, 1): np.ones((3, 3)),
        (0, 1): np.ones((2, 3)),
    })
    prior = gp.prior(raw=True)
    util.assert_equal(prior[0, 1], prior[1, 0].T)

def test_new_proc():
    gp = lgp.GP()
    gp._procs[0] = None
    gp.addx(0, 0, proc=0)
    with pytest.raises(TypeError):
        gp.prior(0, raw=True)

def test_partial_derivative():
    gp = lgp.GP(lgp.ExpQuad())
    x = np.arange(20)
    y = np.zeros(len(x), 'f8,f8')
    y['f0'] = x
    gp.addx(y, 0, deriv='f0')
    cov1 = gp.prior(0, raw=True)
    
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(x, 0, deriv=1)
    cov2 = gp.prior(0, raw=True)
    
    util.assert_equal(cov1, cov2)

def test_zero_covblock():
    gp = lgp.GP()
    a = gen.standard_normal((10, 10))
    m = a.T @ a
    gp.addcov(m, 0)
    gp.addcov(m, 1)
    prior = gp.prior(raw=True)
    util.assert_equal(prior[0, 1], np.zeros_like(m))

def test_addcov_checks():
    a = gen.standard_normal((10, 10))
    b = np.copy(a)
    b[0, 0] = np.inf
    m = b.T @ b

    gp = lgp.GP()
    with pytest.raises(ValueError):
        gp.addcov(a, 0)
    with pytest.raises(ValueError):
        gp.addcov(m, 0)
    
    gp = lgp.GP(checksym=False)
    gp.addcov(a, 0)
    
    gp = lgp.GP(checkfinite=False)
    gp.addcov(m, 0)
    
    a = a @ a.T
    gp = lgp.GP()
    dec = lgp.GP.decompose(a)
    with pytest.raises(TypeError):
        gp.addcov({(0, 0): a}, decomps=dec)
    with pytest.raises(KeyError):
        gp.addcov({(0, 0): a}, decomps={1: dec})
    with pytest.raises(TypeError):
        gp.addcov({(0, 0): a}, decomps={0: a})
    b = gen.standard_normal((20, 20))
    b = b @ b.T
    bd = lgp.GP.decompose(b)
    with pytest.raises(ValueError):
        gp.addcov({(0, 0): a}, decomps={0: bd})

def test_makecovblock_checks():
    a = gen.standard_normal((10, 10))
    b = np.copy(a)
    b[0, 0] = np.inf
    m = b.T @ b
    
    gp = lgp.GP(checksym=False)
    gp.addcov(a, 0)
    gp._checksym = True
    with pytest.raises(RuntimeError):
        gp.prior(raw=True)

    gp = lgp.GP(checkfinite=False)
    gp.addcov(m, 0)
    gp._checkfinite = True
    with pytest.raises(RuntimeError):
        gp.prior(raw=True)
    
    gp = lgp.GP(checksym=False, checkpos=False)
    gp.addcov(a, 0)
    gp.prior(raw=True)
    
    gp = lgp.GP(checkfinite=False, checkpos=False)
    gp.addcov(m, 0)
    gp.prior(raw=True)

def test_covblock_checks():
    a, b, c, d = gen.standard_normal((4, 10, 10))
    m = a.T @ a
    n = b.T @ b
    gp = lgp.GP(checksym=False, checkpos=False)
    gp.addcov({
        (0, 0): m,
        (1, 1): n,
        (0, 1): c,
        (1, 0): d,
    })
    gp2 = copy.deepcopy(gp)
    gp._checksym = True
    with pytest.raises(RuntimeError):
        gp.prior(raw=True)
    gp2.prior(raw=True)

def test_solver_cache():
    gp = lgp.GP(lgp.ExpQuad())
    x = np.linspace(0, 1, 10)
    y = np.linspace(1, 2, 10)
    z = np.zeros_like(x)
    gp.addx(x, 0)
    gp.addx(y, 1)
    m1, c1 = gp.predfromdata({0: z}, 1, raw=True)
    m2, c2 = gp.predfromdata({0: z}, 1, raw=True)
    util.assert_equal(m1, m2)
    util.assert_equal(c1, c2)

def test_checkpos():
    a = gen.standard_normal((20, 20))
    m = a.T @ a
    w, v = np.linalg.eigh(m)
    w[np.arange(len(w)) % 2] *= -1
    m = (v * w) @ v.T
    
    gp = lgp.GP()
    gp.addcov(m, 0)
    with pytest.raises(np.linalg.LinAlgError):
        gp.prior()
    
    gp._checkpositive = False
    gp.prior()

def test_priorpoints_cache():
    gp = lgp.GP(lgp.ExpQuad())
    x = np.arange(20)
    gp.addx(x, 0)
    gp.addx(x, 1)
    prior = gp.prior()
    cov = gvar.evalcov(prior)
    util.assert_equal(cov[0, 0], cov[1, 1])
    util.assert_equal(cov[0, 0], cov[0, 1])

def test_priortransf():
    gp = lgp.GP(lgp.ExpQuad())
    x, y = np.arange(40).reshape(2, -1)
    gp.addx(x, 0)
    gp.addx(y, 1)
    gp.addtransf({0: x, 1: y}, 2)
    cov1 = gp.prior(2, raw=True)
    u = gp.prior(2)
    cov2 = gvar.evalcov(u)
    util.assert_allclose(cov1, cov2, rtol=1e-15)

def test_new_element():
    gp = lgp.GP()
    gp._elements[0] = None
    with pytest.raises(Exception):
        gp.prior()

def test_given_checks():
    gp = lgp.GP(lgp.ExpQuad())
    x, y, z = gen.standard_normal((3, 20))
    gp.addx(x, 0)
    gp.addx(y, 1)
    with pytest.raises(TypeError):
        gp.predfromdata(0, 1)
    with pytest.raises(TypeError):
        gp.predfromdata({0: z}, 1, givencov=0)
    with pytest.raises(KeyError):
        gp.predfromdata({2: z}, 1)
    with pytest.raises(ValueError):
        gp.predfromdata({0: z[:-1]}, 1)
    with pytest.raises(TypeError):
        gp.predfromdata({0: np.empty_like(z, str)}, 1)

def test_zero_givencov():
    gp = lgp.GP(lgp.ExpQuad())
    x, y, z = gen.standard_normal((3, 20))
    gp.addx(x, 0)
    gp.addx(y, 1)
    cov = np.zeros(2 * x.shape)
    m1, c1 = gp.predfromdata({0: z}, 1, {(0, 0): cov}, raw=True)
    m2, c2 = gp.predfromdata({0: z}, 1, raw=True)
    util.assert_equal(m1, m2)
    util.assert_equal(c1, c2)

def test_pred_checks():
    gp = lgp.GP(lgp.ExpQuad())
    x, y, z = gen.standard_normal((3, 20))
    gp.addx(x, 0)
    gp.addx(y, 1)
    with pytest.raises(ValueError):
        gp.pred({0: z}, 1)
    with pytest.raises(ValueError):
        gp.predfromdata({0: z}, 1, raw=True, keepcorr=True)
    with pytest.raises(ValueError):
        gp.predfromdata({0: np.full_like(z, np.nan)}, 1)
    with pytest.raises(ValueError):
        gp.predfromdata({0: z}, 1, {(0, 0): np.full(2 * x.shape, np.nan)})
    a = gen.standard_normal((20, 20))
    with pytest.raises(ValueError):
        gp.predfromdata({0: z}, 1, {(0, 0): a})
    gp._checkfinite = False
    gp.predfromdata({0: np.full_like(z, np.nan)}, 1)

def test_pred_all():
    gp = lgp.GP(lgp.ExpQuad())
    x, y, z = gen.standard_normal((3, 20))
    gp.addx(x, 0)
    gp.addx(y, 1)
    m1, c1 = gp.predfromdata({0: z}, raw=True)
    m2, c2 = gp.predfromdata({0: z}, [0, 1], raw=True)
    util.assert_equal(m1, m2)
    util.assert_equal(c1, c2)

def test_marginal_likelihood_separate():
    gp = lgp.GP(lgp.ExpQuad())
    x, y = gen.standard_normal((2, 20))
    gp.addx(x, 0)
    ml1 = gp.marginal_likelihood({0: y})
    l, r = gp.marginal_likelihood({0: y}, separate=True)
    ml2 = -1/2 * (l + r @ r)
    util.assert_allclose(ml2, ml1, rtol=1e-15)

def test_marginal_likelihood_checks():
    gp = lgp.GP(lgp.ExpQuad())
    x, y = gen.standard_normal((2, 20))
    gp.addx(x, 0)
    z = np.full_like(x, np.nan)
    with pytest.raises(ValueError):
        gp.marginal_likelihood({0: z})
    m = np.full(2 * x.shape, np.nan)
    with pytest.raises(ValueError):
        gp.marginal_likelihood({0: y}, {(0, 0): m})
    a = gen.standard_normal(2 * x.shape)
    with pytest.raises(ValueError):
        gp.marginal_likelihood({0: y}, {(0, 0): a})
    c = a.T @ a
    with pytest.warns(UserWarning):
        gp.marginal_likelihood({0: gvar.gvar(y, c)}, {(0, 0): c})

def test_marginal_likelihood_gvar():
    gp = lgp.GP(lgp.ExpQuad())
    x, y = gen.standard_normal((2, 20))
    gp.addx(x, 0)
    a = gen.standard_normal((20, 20))
    m = a.T @ a
    ml1 = gp.marginal_likelihood({0: gvar.gvar(y, m)})
    ml2 = gp.marginal_likelihood({0: y}, {(0, 0): m})
    util.assert_allclose(ml2, ml1, rtol=1e-15)

def test_singleton():
    dp = lgp.GP.DefaultProcess
    assert repr(dp) == 'DefaultProcess'
    with pytest.raises(NotImplementedError):
        dp()

def test_addtransf_abstract():
    def func():
        gp = lgp.GP(lgp.ExpQuad())
        gp.addx(0, 0)
        gp.addtransf({0: np.inf}, 1)
        return gp.prior(1, raw=True)
    with pytest.raises(ValueError):
        func()
    assert jax.jit(func)().item() == np.inf

def test_addlintransf_abstract():
    def func():
        gp = lgp.GP(lgp.ExpQuad())
        gp.addx(0, 0)
        gp.addlintransf(lambda x: x + 1, [0], 1)
        return gp.prior(1, raw=True)
    with pytest.raises(RuntimeError):
        func()
    assert jax.jit(func)().item() == 3

def test_addcov_abstract():
    def func():
        gp = lgp.GP(lgp.ExpQuad())
        gp.addcov({
            (0, 0): 1,
            (1, 1): 1,
            (0, 1): 1,
            (1, 0): 0,
        })
        return gp.prior([0, 1], raw=True)
    with pytest.raises(ValueError):
        func()
    cov = jax.jit(func)()
    assert cov[0, 1] == 1 or cov[0, 1] == 0

def test_marginal_likelihood_abstract():
    def func():
        gp = lgp.GP(lgp.ExpQuad())
        gp.addx(gen.standard_normal(10), 0)
        return gp.marginal_likelihood({0: np.full(10, np.nan)})
    
    with pytest.raises(ValueError):
        func()
    assert np.isnan(jax.jit(func)())
    
    def func(cov):
        gp = lgp.GP(lgp.ExpQuad())
        gp.addx(gen.standard_normal(10), 0)
        return gp.marginal_likelihood({0: gen.standard_normal(10)}, {(0, 0): cov})
    
    covnan = np.full((10, 10), np.nan)
    with pytest.raises(ValueError):
        func(covnan)
    assert np.isnan(jax.jit(func)(covnan))
    
    covasym = gen.standard_normal((10, 10))
    with pytest.raises(ValueError):
        func(covasym)
    jax.jit(func)(covasym)

def test_addcov_decomps():
    a = gen.standard_normal((10, 10))
    a = a @ a.T
    blocks = {
        (0, 0): a[:5, :5],
        (0, 1): a[:5, 5:],
        (1, 0): a[5:, :5],
        (1, 1): a[5:, 5:],
    }
    dec = lgp.GP.decompose(blocks[0, 0])
    dec1 = lgp.GP.decompose(blocks[1, 1])
    b = jnp.asarray(gen.standard_normal(5))
    
    def makez(**kw):
        gp = lgp.GP()
        gp.addcov(blocks, **kw)
        return gp.predfromdata({0: b}, 1)
    
    z1 = makez()
    z2 = makez(decomps={0: dec})
    z3 = makez(decomps={0: dec1})
    
    util.assert_similar_gvars(z1, z2)
    with pytest.raises(AssertionError):
        util.assert_similar_gvars(z1, z3)
    
    def makez(**kw):
        gp = lgp.GP()
        gp.addcov(blocks[0, 0], 0, **kw)
        gp.addcov(blocks[1, 1], 1)
        return gp.predfromdata({0: b}, 1)
    
    z1 = makez()
    z2 = makez(decomps=dec)
    
    util.assert_similar_gvars(z1, z2)

def test_matrices():
    shapes = [
        (),
        (10,),
        (3, 7),
    ]
    for sout in shapes:
        tensors = []
        gp = lgp.GP(lgp.ExpQuad())
        for i, sin in enumerate(shapes):
            tensor = gen.standard_normal(sout + sin)
            x = gen.standard_normal(sin)
            gp.addx(x, i)
            tensors.append(tensor)
            def transf(*args):
                out = 0
                for x, t, s in zip(args, tensors, shapes):
                    out += jnp.tensordot(t, x, axes=len(s))
                return out
            gp.addlintransf(transf, list(range(len(tensors))), 100 + i)
            
            matrices = gp._elements[100 + i].matrices(gp)
            tensors2 = [m.reshape(t.shape) for m, t in zip(matrices, tensors)]
            for t, t2 in zip(tensors, tensors2):
                util.assert_equal(t2, t)

def test_transf_outer():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(np.arange(5), 0)
    t = np.arange(5)
    gp.addtransf({0: t}, 1, axes=0)
    cov = gp.prior(raw=True)
    c0 = cov[0, 0]
    c1 = np.einsum('i,jl,k', t, c0, t)
    util.assert_equal(c1, cov[1, 1])

def test_transf_checks():
    gp = lgp.GP(lgp.ExpQuad())
    with pytest.raises(ValueError):
        gp.addtransf({}, 2)

def test_givencov_decomp():

    def genpd(n, rank=None, size=()):
        if not isinstance(size, tuple):
            size = (size,)
        if rank is None:
            rank = n
        m = gen.standard_normal(size + (n, rank))
        return m @ np.swapaxes(m, -2, -1)
    
    def decs(gp, keys, covrank=None):
        elems = gp._elements
        shapes = [elems[key].shape for key in keys]
        given = {k: np.zeros(s) for k, s in zip(keys, shapes)}
        size = sum(elems[key].size for key in keys)
        cov = genpd(size, covrank)
        slices = gp._slices(keys)
        givencov1 = {
            (ka, kb): cov[sla, slb].reshape(sa + sb)
            for (ka, sla, sa), (kb, slb, sb)
            in itertools.product(zip(keys, slices, shapes), repeat=2)
        }
        givencov2 = gp.decompose(cov)
        dec1, _ = gp._prior_decomp(given, givencov1)
        dec2, _ = gp._prior_decomp(given, givencov2)
        classes = (lgp._linalg.Woodbury, lgp._linalg.Woodbury2)
        assert not isinstance(dec1, classes)
        assert isinstance(dec2, classes)
        return dec1, dec2
    
    # generic matrix
    a = genpd(10)
    gp = lgp.GP()
    gp.addcov(a, 0)
    dec1, dec2 = decs(gp, [0])
    util.assert_close_decomps(dec2, dec1, rtol=1e-11)
    
    # short sandwich
    b = gen.standard_normal((len(a) // 2, len(a)))
    gp.addtransf({0: b}, 1)
    dec1, dec2 = decs(gp, [1])
    util.assert_close_decomps(dec2, dec1, rtol=1e-7)
    
    # tall sandwich
    c = gen.standard_normal((len(a) * 2, len(a)))
    gp.addtransf({0: c}, 2)
    dec1, dec2 = decs(gp, [2])
    util.assert_close_decomps(dec2, dec1, rtol=1e-9)
    
    # short and tall sandwich
    dec1, dec2 = decs(gp, [1, 2])
    util.assert_close_decomps(dec2, dec1, rtol=1e-10)

    # two generic matrices
    d = genpd(20)
    gp.addcov(d, 3)
    dec1, dec2 = decs(gp, [0, 3])
    util.assert_close_decomps(dec2, dec1, rtol=1e-8)
    
    # matrix, short and tall sandwich
    dec1, dec2 = decs(gp, [0, 1, 2])
    util.assert_close_decomps(dec2, dec1, rtol=1e-7)
    
    # short and tall sandwich, starting from different matrices
    e = gen.standard_normal((2 * len(d), len(d)))
    gp.addtransf({3: e}, 4)
    dec1, dec2 = decs(gp, [1, 4])
    util.assert_close_decomps(dec2, dec1, rtol=1e-9)
    
    # sum of two matrices
    f = genpd(len(a))
    gp.addcov(f, 5)
    gp.addtransf({0: 1, 5: 1}, 6)
    dec1, dec2 = decs(gp, [6])
    util.assert_close_decomps(dec2, dec1, rtol=1e-11)
    
    # the same matrix, twice
    gp.addcov({(k, q): a for k in [7, 8] for q in [7, 8]})
    dec1, dec2 = decs(gp, [7, 8])
    util.assert_close_decomps(dec2, dec1, rtol=1e-10)
    
    # low rank givencov
    dec1, dec2 = decs(gp, [0], len(a) // 2)
    util.assert_close_decomps(dec2, dec1, rtol=1) # TODO wildly inaccurate
    
def test_nochecksym_structured():
    gp = lgp.GP(lgp.ExpQuad(), checksym=False, halfmatrix=True)
    gp.addx(np.zeros(1, 'd,d'), 0)
    gp.prior(0, raw=True)

def test_nochecksym_structured_jit():
    jax.jit(test_nochecksym_structured)()

def test_nochecksym_tracer():
    def fun():
        gp = lgp.GP(lgp.ExpQuad(), checksym=False, halfmatrix=True)
        gp.addx(np.zeros(1), 0)
        return gp.prior(0, raw=True)
    jax.jit(fun)()

def test_decompose_nd():
    cov = np.array(2)
    d1 = lgp.GP.decompose(cov)
    d2 = lgp.GP.decompose(cov.reshape(1, 1))
    d3 = lgp.GP.decompose(cov.reshape(1, 1, 1, 1))
    util.assert_close_decomps(d1, d2)
    util.assert_close_decomps(d1, d3)

def test_pred_woodbury():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(0, 0)
    gp.addx(1, 1)
    cov = 2
    covdec = gp.decompose(cov)
    y1 = gp.predfromdata({0: 1}, 1, {(0, 0): cov})
    y2 = gp.predfromdata({0: 1}, 1, covdec)
    util.assert_similar_gvars(y1, y2, rtol=1e-15)

def test_pred_ambiguous_error_covariance():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(0, 0)
    gp.addx(1, 1)
    with pytest.raises(ValueError):
        gp.predfromdata({0: gvar.gvar(0, 1)}, 1, {(0, 0): 2})

def test_pred_gvars_givencov():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(0, 0)
    gp.addx(1, 1)
    mean, sdev = 1, 2
    y1 = gp.predfromdata({0: gvar.gvar(mean, sdev)}, 1)
    y2 = gp.predfromdata({0: mean}, 1, {(0, 0): sdev ** 2})
    y3 = gp.predfromdata({0: mean}, 1, gp.decompose(sdev ** 2))
    util.assert_similar_gvars(y1, y2)
    util.assert_similar_gvars(y1, y3)

def test_pred_fromfit_decomp():
    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(0, 0)
    gp.addx(1, 1)
    mean, sdev = 1, 2
    y0 = gp.predfromdata({0: gvar.gvar(mean, sdev)}, 0)
    y1 = gp.predfromfit({0: y0}, 1, keepcorr=False)
    y2 = gp.predfromfit({0: gvar.mean(y0)}, 1, {(0, 0): gvar.var(y0)}, keepcorr=False)
    y3 = gp.predfromfit({0: gvar.mean(y0)}, 1, gp.decompose(gvar.var(y0)), keepcorr=False)
    util.assert_similar_gvars(y1, y2)
    util.assert_similar_gvars(y1, y3)



"""Fit of parton distributions functions (PDFs)"""

import lsqfitgp as lgp
import numpy as np
from matplotlib import pyplot as plt
import gvar

np.random.seed(20220416)

#### DEFINE MODEL ####
# for each gluon:
# h ~ GP
# f = h''
# int_0^1 dx f(x) = [h'(x)]_0^1 = h'(1) - h'(0)
# int_0^1 dx x f(x) = [xh'(x) - h(x)]_0^1 = h'(1) - h(1) + h(0)

xtype = np.dtype([
    ('x'    , float),
    ('gluon', int  ),
])

kernel = lgp.ExpQuad(dim='x') * lgp.White(dim='gluon')

xdata = np.empty((8, 30), xtype)
xdata['gluon'] = np.arange(8)[:, None]
xdata[    'x'] = np.linspace(0, 1, 30)
    
M = np.random.randn(20, 8, 30) # xdata -> data transform

xinteg = np.empty((8, 2), xtype)
xinteg['gluon'] = np.arange(8)[:, None]
xinteg[    'x'] = [0, 1]

suminteg = np.empty(xinteg.shape)
suminteg[:, 0] = -1
suminteg[:, 1] =  1

xintegx = np.empty(8, xtype)
xintegx['gluon'] = np.arange(8)
xintegx[    'x'] = 1

#### CREATE GP OBJECT ####

gp = lgp.GP(kernel)

gp.addx(xdata, 'xdata', deriv=(2, 'x'))
gp.addtransf({'xdata': M}, 'data', axes=2)

gp.addx(xinteg, 'xinteg', deriv='x')
gp.addtransf({'xinteg': suminteg}, 'suminteg', axes=2)

gp.addx(xinteg, 'xintegx0')
gp.addx(xintegx[None], 'xintegx1', deriv='x')
gp.addtransf({'xintegx1': np.ones((1, 8)), 'xintegx0': -suminteg}, 'sumintegx', axes=2)

#### GENERATE FAKE DATA ####

prior = gp.predfromdata({
    'suminteg' : 1,
    'sumintegx': 1,
}, ['data', 'xdata'])
priorsample = next(gvar.raniter(prior))

datamean = priorsample['data']
dataerr = np.full_like(datamean, 1)
datamean = datamean + dataerr * np.random.randn(*dataerr.shape)
data = gvar.gvar(datamean, dataerr)

# check the integral is one with trapezoid rule
x = xdata['x']
y = priorsample['xdata']
checksum = np.sum((      y[:, 1:] +       y[:, :-1]) / 2 * np.diff(x, axis=1))
print('sum_i int dx   f_i(x) =', checksum)
checksum = np.sum(((y * x)[:, 1:] + (y * x)[:, :-1]) / 2 * np.diff(x, axis=1))
print('sum_i int dx x f_i(x) =', checksum)

#### FIT ####

pred = gp.predfromdata({
    'suminteg' :    1,
    'sumintegx':    1,
    'data'     : data,
}, ['data', 'xdata'])

# check the integral is one with trapezoid rule
x = xdata['x']
y = pred['xdata']
checksum = np.sum((      y[:, 1:] +       y[:, :-1]) / 2 * np.diff(x, axis=1))
print('sum_i int dx   f_i(x) =', checksum)
checksum = np.sum(((y * x)[:, 1:] + (y * x)[:, :-1]) / 2 * np.diff(x, axis=1))
print('sum_i int dx x f_i(x) =', checksum)

#### PLOT RESULTS ####

fig, axs = plt.subplots(1, 2, num='pdf1', clear=True, figsize=[9, 4.5])
axs[0].set_title('PDFs')
axs[1].set_title('Data')

for i in range(len(xdata)):
    y = pred['xdata'][i]
    m = gvar.mean(y)
    s = gvar.sdev(y)
    axs[0].fill_between(xdata[i]['x'], m - s, m + s, alpha=0.6, facecolor=f'C{i}')
    y2 = priorsample['xdata'][i]
    axs[0].plot(xdata[i]['x'], y2, color=f'C{i}')

m = gvar.mean(pred['data'])
s = gvar.sdev(pred['data'])
x = np.arange(len(data))
axs[1].fill_between(x, m - s, m + s, step='mid', color='lightgray')
axs[1].errorbar(x, datamean, dataerr, color='black', linestyle='', capsize=2)

fig.tight_layout()
fig.show()



# lsqfitgp/_patch_jax/__init__.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

# TODO disable this jax 0.4.4 feature because it makes empbayes_fit fail due to
# leaked tracers, but I can't fix it and maybe it's a jax bug. Wait till next
# release to see what happens. => maybe it's
# https://github.com/google/jax/issues/14776, "jax.ensure_compile_time_eval puts
# the trace stack in a bad state"? => still not working in 0.4.8
import os
os.environ['JAX_JIT_PJIT_API_MERGE'] = '0'

from jax.config import config
config.update("jax_enable_x64", True)

import traceback
import functools

import jax
from jax import core
from jax import lax
from jax import numpy as jnp
from jax import tree_util

from ._batcher import batchufunc
from ._fasthash import fasthash64, fasthash32

def makejaxufunc(ufunc, *derivs):
    """
    
    Wrap a numpy ufunc to add jax support.
    
    Parameters
    ----------
    ufunc : callable
        Elementwise function following numpy broadcasting and type promotion
        rules. Keyword arguments not supported.
    derivs : sequence of callable
        Derivatives of the function w.r.t. each positional argument, with the
        same signature as `ufunc`. Pass None to indicate a missing derivative.
        There must be as many derivatives as the arguments to `ufunc`.
    
    Return
    ------
    func : callable
        Wrapped `ufunc`. Supports jit, but the calculation is performed on cpu.
    
    """

    nondiff_argnums = [i for i, d in enumerate(derivs) if d is None]
    
    @functools.wraps(ufunc)
    @functools.partial(jax.custom_jvp, nondiff_argnums=nondiff_argnums)
    def func(*args):
        args = tuple(map(jnp.asarray, args))
        class ResultDummy:
            shape = jnp.broadcast_shapes(*(arg.shape for arg in args))
            dtype = jnp.result_type(*args)
        return jax.pure_callback(ufunc, ResultDummy, *args, vectorized=True)

    @func.defjvp
    def func_jvp(*allargs):
        ndargs = allargs[:-2]
        dargs = allargs[-2]
        dargst = allargs[-1]
        
        itnd = iter(ndargs)
        itd = iter(dargs)
        args = [next(itnd) if d is None else next(itd) for d in derivs]
        
        result = func(*args)
        tangent = sum([
            d(*args) * t for t, d in
            zip(dargst, (d for d in derivs if d is not None))
        ])
        return result, tangent
    
    return func

def elementwise_grad(fun, argnum=0):
    assert int(argnum) == argnum and argnum >= 0, argnum
    @functools.wraps(fun)
    def funderiv(*args):
        preargs = args[:argnum]
        postargs = args[argnum + 1:]
        def oneargfun(arg):
            args = preargs + (arg,) + postargs
            return fun(*args)
        primal = args[argnum]
        shape = getattr(primal, 'shape', ())
        dtype = getattr(primal, 'dtype', type(primal))
        tangent = jnp.ones(shape, dtype)
        primal_out, tangent_out = jax.jvp(oneargfun, (primal,), (tangent,))
        return tangent_out
    return funderiv

class skipifabstract:
    """
    Context manager to try to do all operations eagerly even during jit, and
    skip entirely if it is not possible.
    """
    # I feared this would be slow because of the slow jax exception handling,
    # but %timeit suggests it isn't

    ENSURE_COMPILE_TIME_EVAL = True
    ENABLED = True
    
    def __enter__(self):
        if self.ENSURE_COMPILE_TIME_EVAL and self.ENABLED:
            self.mgr = jax.ensure_compile_time_eval()
            self.mgr.__enter__()
    
    def __exit__(self, exc_type, exc_value, tb):
        if not self.ENABLED:
            return
        exit = None
        if self.ENSURE_COMPILE_TIME_EVAL:
            exit = self.mgr.__exit__(exc_type, exc_value, tb)
        if exit or exc_type in (jax.errors.ConcretizationTypeError, jax.errors.TracerArrayConversionError):
            return True
        
        weird_cond = exc_type is IndexError and (
            traceback.extract_tb(tb)[-1].name in ('arg_info_pytree', '_origin_msg'),
        )
        if weird_cond: # pragma: no cover
            # TODO this ignores a jax internal bug I don't understand, appears
            # in examples/pdf4.py
            return True

# TODO make stop_hessian work in reverse mode
# see jax issue #10994
# from jax.interpreters import ad
# from jax._src import ad_util
# ad.primitive_transposes[ad_util.stop_gradient_p] = lambda ct, _: [lax.stop_gradient(ct)]
# try with jax.jvp and lax.stop_gradient, I tried once and I failed but I did
# not try hard enough

@jax.custom_jvp
def stop_hessian(x):
    return x

@stop_hessian.defjvp
def stop_hessian_jvp(primals, tangents):
    x, = primals
    x_dot, = tangents
    return x, lax.stop_gradient(x_dot)

def value_and_ops(f, *ops, has_aux=False, **kw):
    """
    Creates a function returning the values of f and its derivatives defined
    by stacking the operators in ops. Example:
    value_and_grad_and_hess = value_and_ops(f, jax.grad, jax.jacfwd)
    """
    if not ops:
        return f
    def fop(*args, **kw):
        y = f(*args, **kw)
        if has_aux:
            y, aux = y
            return y, (aux,)
        else:
            return y, ()
    def nextfop(fop):
        def nextfop(*args, **kw):
            y, aux = fop(*args, **kw)
            return y, aux + (y,)
        return nextfop
    for op in ops:
        fop = op(nextfop(fop), has_aux=True, **kw)
    @functools.wraps(f)
    def lastfop(*args, **kw):
        y, aux = fop(*args, **kw)
        if has_aux:
            return aux[1:] + (y,), aux[0]
        else:
            return aux + (y,)
    return lastfop

def float_type(*args):
    t = jnp.result_type(*args)
    return jnp.sin(jnp.empty(0, t)).dtype



import functools

import jax
from jax import numpy as jnp

class Cippa:
    
    @classmethod
    @functools.partial(jax.jit, static_argnums=(0,))
    def lippa(cls, x):
        try:
            if jnp.all(x):
                print('ciao')
        except jax.errors.ConcretizationTypeError:
            pass
        else:
            raise RuntimeError('x is static')
        return x + 1

print(Cippa.lippa(4))
print(Cippa.lippa(jnp.arange(10)))



# lsqfitgp/examples/t.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE T.

    Where we pretend to discover that two series of events
    were in fact one the delayed and imperfect copy of the
    other.

"""

import lsqfitgp as lgp
from matplotlib import pyplot as plt
from jax import numpy as jnp
import numpy as np
import gvar

time = np.arange(21)
delay = 20

def makex(time, delay):
    x = np.empty((2, time.size), dtype=[
        ('time', float),
        ('label', int)
    ])
    x['time'][0] = time
    x['time'][1] = time - delay
    x['label'][0] = 0
    x['label'][1] = 1
    return lgp.StructuredArray(x)

x = makex(time, delay)

label_names = ['gatti_comprati', 'gatti_morti']

function = lambda x: np.exp(-1/2 * ((x - 10) / 5)**2)
data_mean = function(x['time'])
data_mean[1] += 0.02 * time
data = gvar.gvar(data_mean, np.full_like(data_mean, 0.05))
data = gvar.make_fake_data(data)

def makegp(params):
    kernel = lgp.Cauchy(scale=params['time_scale'], dim='time', beta=2)
    kernel *= lgp.ExpQuad(scale=params['label_scale'], dim='label')
    gp = lgp.GP(kernel)
    xmod = x.at['time'].set(jnp.array([time, time - params['delay']]))
    gp.addx(xmod, 'A')
    return gp

hyperprior = gvar.BufferDict({
    'log(time_scale)': gvar.log(gvar.gvar(10, 10)),
    'log(label_scale)': gvar.log(gvar.gvar(5, 5)),
    'delay': gvar.gvar(5, 20)
})

fit = lgp.empbayes_fit(hyperprior, makegp, {'A': data}, raises=False, jit=True, verbosity=2)

print(f'time scale = {fit.p["time_scale"]}')
corr = lgp.ExpQuad(scale=gvar.mean(fit.p['label_scale']))(0, 1)
print(f'correlation = {corr:.3g} (equiv. scale = {fit.p["label_scale"]})')
print(f'delay = {fit.p["delay"]} (true = {delay})')

fig, ax = plt.subplots(num='t', clear=True)

time_pred = np.linspace(np.min(time), np.max(time) + 1.5 * (np.max(time) - np.min(time)), 100)

for style, params_sample in zip(['-', '--'], gvar.raniter(fit.p, 2)):
    
    gp = makegp(params_sample)
    xpred = makex(time_pred, params_sample['delay'])
    gp.addx(xpred, 'B')
    pred = gp.predfromdata({'A': data}, 'B')

    for sample in gvar.raniter(pred, 1):
        for i in range(2):
            label = f'{label_names[i]}, delay={params_sample["delay"]:.1f}'
            ax.plot(time_pred, sample[i], color=f'C{i}', alpha=0.5, label=label, linestyle=style)

for i in range(2):
    ax.errorbar(time, gvar.mean(data[i]), yerr=gvar.sdev(data[i]), fmt='.', color=f'C{i}')

ax.legend(loc='best')
ax.set_xlabel('time')

fig.show()



import numpy as np
from matplotlib import pyplot as plt
import mpmath

import lsqfitgp as lgp

zeta = np.vectorize(lambda *args: float(mpmath.zeta(*args)))

def dlmf_25_11_7(s, a):
    # for integer s <= 0
    return a ** -s + (1 + a) ** -s * (1/2 + (1 + a) / (s - 1))

fig, ax = plt.subplots(num='zeta', clear=True)

a = np.linspace(0, 1, 1000)
for s in np.arange(0, -10, -1):
    z = dlmf_25_11_7(s, a)
    zsure = zeta(s, a)
    line, = ax.plot(a, zsure, label=f's={s}')
    ax.plot(a, z, color=line.get_color(), linestyle='--')

ax.legend()

fig.show()



# lsqfitgp/examples/runexamples.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""Run the scripts given on the command line and saves the figures produced
in the same directory of each corresponding script."""

import sys
import os
import warnings

import numpy as np
from matplotlib import pyplot as plt
import gvar

sys.path.insert(0, '..') # to import lsqfitgp if run from examples/
sys.path.insert(0, '.')

warnings.filterwarnings('ignore', r'Matplotlib is currently using agg, which is a non-GUI backend, so cannot show the figure\.')

for file in sys.argv[1:]:
    print('\nrunexamples.py: running {}...'.format(file))
    with open(file, 'r') as stream:
        code = stream.read()
    
    # reset working environment
    plt.close('all')
    np.random.seed(0)
    gvar.switch_gvar()

    globals_dict = {}
    exec(code, globals_dict)
    
    # save figures
    nums = plt.get_fignums()
    path, name = os.path.split(file)
    base, _ = os.path.splitext(name)
    prefix = os.path.join(path, 'plot', base)
    directory, _ = os.path.split(prefix)
    os.makedirs(directory, exist_ok=True)
    for num in nums:
        fig = plt.figure(num)
        suffix = f'-{num}' if num > 1 else ''
        out = f'{prefix}{suffix}.png'
        print(f'runexamples.py: write {out}')
        fig.savefig(out)



import jax

ArrayMeta = type(jax.numpy.ndarray)

class JaxArray(metaclass=ArrayMeta): pass

class ndarray(jax.numpy.ndarray, JaxArray):
    """replacement for jax.numpy.ndarray"""
    pass

jax.numpy.ndarray = ndarray
jax._src.numpy.lax_numpy.ndarray = ndarray

x = jax.numpy.zeros(5)

print(isinstance(x, JaxArray))
print(issubclass(type(x), JaxArray)) # -> False, not really working



import numpy as np
from scipy import special
from matplotlib import pyplot as plt

import lsqfitgp as lgp

fig, ax = plt.subplots(num='exp1', clear=True)

x = np.linspace(1e-3, 20, 1000)
y1 = lgp._patch_jax.exp1_imag(x)
y2 = special.exp1(-1j * x)
ax.plot(x, y1.real, label=f'lgp real')
ax.plot(x, y2.real, label=f'scipy real', linestyle='--')
ax.plot(x, y1.imag, label=f'lgp imag')
ax.plot(x, y2.imag, label=f'scipy imag', linestyle='--')
ax.legend()

fig.show()



import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np
import gvar

fig, ax = plt.subplots(num='kernel', clear=True)

kernels = [
    # ['expquad', lgp.ExpQuad()],
    # ['cos', lgp.ExpQuad(scale=3) * lgp.Cos()],
    # ['wiener', lgp.Wiener()],
    # ['fb1/2', lgp.FracBrownian()],
    # ['fb1/10', lgp.FracBrownian(H=1/10)],
    # ['fb9/10', lgp.FracBrownian(H=9/10)],
    # ['fb0.99', lgp.FracBrownian(H=99/100)],
    # ['NN', lgp.NNKernel(loc=10)],
    # ['Fourier(n=1)', lgp.Fourier(n=1, scale=10)],
    # ['Fourier(n=2)', lgp.Fourier(n=2, scale=10)],
    # ['Celerite', lgp.Celerite(gamma=0.1)],
    # ['Harmonic', lgp.Harmonic(Q=10)],
    # ['BrownianBridge', lgp.BrownianBridge()],
    # ['Taylor', lgp.Taylor()],
    ['Pink', lgp.Pink(dw=1e20)],
]

for label, kernel in kernels:
    gp = lgp.GP(kernel)
    x = np.linspace(0, 20, 2000)
    gp.addx(x, 'x', deriv=0)
    y = gp.prior('x')
    samples = np.transpose(list(gvar.raniter(y, 1, eps=1e-12)))
    ax.plot(x, samples, label=label)

ax.legend(loc='best')
fig.show()



import jax
from jax import numpy as jnp

with jax.checking_leaks():
    @jax.jit
    @jax.grad
    def test(xy):
        x, y = xy
        try:
            with jax.ensure_compile_time_eval():
                assert jnp.all(x >= 0)
                assert jnp.all(y >= 0)
        except jax.errors.ConcretizationTypeError:
            pass
        # with jax.ensure_compile_time_eval():
        #     assert jnp.all(x >= 0)
        #     assert jnp.all(y >= 0)
        return x @ y
    test((jnp.arange(8.), jnp.arange(8.)))



# lsqfitgp/tests/test_raniter.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys

import numpy as np
from scipy import stats
import pytest
import gvar

sys.path = ['.'] + sys.path
import lsqfitgp as lgp

def make_mean_cov(n):
    mean = np.random.randn(n)
    a = np.random.randn(n, n)
    cov = a.T @ a
    return mean, cov

def make_mean_cov_dict(*shapes):
    sizes = [np.prod(tuple(s), dtype=int) for s in shapes]
    totsize = sum(sizes)
    mean, cov = make_mean_cov(totsize)
    cumsize = np.cumsum(np.pad(sizes, (1, 0)))
    mean = {
        i: mean[cumsize[i]:cumsize[i + 1]].reshape(shapes[i])
        for i in range(len(shapes))
    }
    cov = {
        (i, j): cov[
            cumsize[i]:cumsize[i + 1],
            cumsize[j]:cumsize[j + 1],
        ].reshape(shapes[i] + shapes[j])
        for i in range(len(shapes))
        for j in range(len(shapes))
    }
    return mean, cov

def test_raniter_randomness():
    n = 40
    mean, cov = make_mean_cov(n)
    samples = list(lgp.raniter(mean, cov, 10 * n))
    smean = np.mean(samples, axis=0)
    scov = np.cov(samples, rowvar=False, ddof=1)
    w, v = np.linalg.eigh(scov / len(samples))
    vsm = v.T @ (smean - mean)
    eps = n * 1e-12 * np.max(w)
    q = (vsm.T / np.maximum(w, eps)) @ vsm
    assert stats.chi2(n).sf(q) > 1e-5
    assert stats.chi2(n).cdf(q) > 1e-5

def test_raniter_packing():
    mean, cov = make_mean_cov(3 + 4)
    dmean = {'a': mean[:3], 'b': mean[3:]}
    dcov = {
        ('a', 'a'): cov[:3, :3],
        ('a', 'b'): cov[:3, 3:],
        ('b', 'a'): cov[3:, :3],
        ('b', 'b'): cov[3:, 3:]
    }
    np.random.seed(0)
    s1 = next(lgp.raniter(mean, cov))
    np.random.seed(0)
    s2 = next(lgp.raniter(dmean, dcov))
    assert np.array_equal(s1, s2.buf)

def test_raniter_warning():
    cov = np.array([1, 1.1, 1.1, 1]).reshape(2, 2)
    with pytest.warns(UserWarning, match='positive definite'):
        next(lgp.raniter(np.zeros(len(cov)), cov))
    with pytest.warns(None) as record:
        next(lgp.raniter(np.zeros(len(cov)), cov, eps=0.1))
    assert len(record) == 0

def test_raniter_shape():
    shape = (2, 5)
    size = np.prod(shape)
    mean, cov = make_mean_cov(size)
    mean = mean.reshape(shape)
    cov = cov.reshape(2 * shape)
    sample = lgp.sample(mean, cov)
    assert sample.shape == shape

def assert_equal_dict_shapes(a, b):
    for k in a:
        assert a[k].shape == b[k].shape
    for k in b:
        assert k in a

def test_raniter_shape_dict():
    mean, cov = make_mean_cov_dict((2, 5), (13,))
    sample = lgp.sample(mean, cov)
    assert_equal_dict_shapes(sample, mean)

def test_raniter_bd():
    mean, cov = make_mean_cov_dict((1,))
    mean = gvar.BufferDict(mean)
    sample = lgp.sample(mean, cov)



from jax import numpy as jnp
from matplotlib import pyplot as plt

def sincder1(x):
    return (x * jnp.cos(x) - jnp.sin(x)) / x ** 2

def sincder2(x):
    a = jnp.expm1(-2j * x + jnp.log1p(1j * x) - jnp.log1p(-1j * x))
    b = jnp.exp(1j * x) * (1 - 1j * x) / (2j * x ** 2)
    return jnp.real(a * b)
    
xtrue = jnp.logspace(-16, 0, 1000)
ytrue = sincder2(xtrue)
x = xtrue.astype('f')
y1 = sincder1(x)
y2 = sincder2(x)

fig, ax = plt.subplots(num='sincder', clear=True)

ax.plot(x, jnp.abs((y1 - ytrue) / ytrue))
ax.plot(x, jnp.abs((y2 - ytrue) / ytrue))
ax.set_xscale('log')
ax.set_yscale('log')

fig.show()



# lsqfitgp/_linalg/_seqalg.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import abc

from jax import numpy as jnp
from jax import lax
from jax import tree_util

from . import _pytree

class SequentialOperation(_pytree.AutoPyTree, metaclass=abc.ABCMeta):
    """see jax.lax.fori_loop for semantics"""
    
    @abc.abstractmethod
    def __init__(self, *args): # pragma: no cover
        pass
    
    @property
    @abc.abstractmethod
    def inputs(self): # pragma: no cover
        """tuple of indices of other ops to be used as input"""
        pass
    
    @abc.abstractmethod
    def init(self, n, *inputs): # pragma: no cover
        """called before the cycle starts with the requested inputs"""
        pass
    
    @abc.abstractmethod
    def iter_out(self, i): # pragma: no cover
        """output passed to other ops who request it through `inputs`,
        guaranteed to be called after `init`"""
        pass
    
    @abc.abstractmethod
    def iter(self, i, *inputs): # pragma: no cover
        """update for iteration"""
        pass

    @abc.abstractmethod
    def finalize(self): # pragma: no cover
        """return final product"""
        pass

def sequential_algorithm(n, ops):
    """
    Define and execute a sequential algorithm on matrices.

    Parameters
    ----------
    n : int
        Number of steps of the algorithm.
    ops : list of SequentialOperation
        Instantiated `SequentialOperation`s that represent the algorithm.

    Return
    ------
    results : tuple
        The sequence of final outputs of each operation in `ops`.
    """
    for i, op in enumerate(ops):
        inputs = op.inputs
        if any(j >= i for j in inputs):
            raise ValueError(f'{i}-th operation {op.__class__.__name__} requested inputs {inputs!r} with forward references')
        args = (ops[j].iter_out(0) for j in inputs)
        op.init(n, *args)
    def body_fun(i, ops):
        for op in ops:
            args = (ops[j].iter_out(i) for j in op.inputs)
            op.iter(i, *args)
        return ops
    ops = lax.fori_loop(1, n, body_fun, ops) # TODO convert to lax.scan and unroll?
    return tuple(op.finalize() for op in ops)
    
class Producer(SequentialOperation):
    """produces something at each iteration but no final output"""
    
    def finalize(self):
        pass

class Consumer(SequentialOperation):
    """produces a final output but no iteration output"""
    
    iter_out = NotImplemented

class SingleInput(SequentialOperation):
    
    def __init__(self, input):
        self.inputs = (input,)
        
    inputs = NotImplemented

class Stack(Consumer, SingleInput):
    """input = an operation producing arrays"""
        
    def init(self, n, a0):
        out = jnp.zeros((n,) + a0.shape, a0.dtype)
        self.out = out.at[0, ...].set(a0)
    
    def iter(self, i, ai):
        self.out = self.out.at[i, ...].set(ai)
    
    def finalize(self):
        """the stacked arrays"""
        return self.out

class MatMulIterByFull(Consumer, SingleInput):
    
    def __init__(self, input, b):
        """input = an operation producing pieces of left operand (a)
        b = right operand"""
        self.inputs = (input,)
        b = jnp.asarray(b)
        assert b.ndim in (1, 2)
        vec = b.ndim < 2
        if vec:
            b = b[:, None]
        self.vec = vec
        self.b = b
    
    @abc.abstractmethod
    def init(self, n, a0): # pragma: no cover
        self.ab = ...
    
    @abc.abstractmethod
    def iter(self, i, ai): # pragma: no cover
        self.ab = ...
    
    def finalize(self):
        ab = self.ab
        if self.vec:
            ab = jnp.squeeze(ab, -1)
        return ab

class MatMulRowByFull(Producer, MatMulIterByFull):
    
    def init(self, n, a0):
        assert a0.ndim == 1
        assert self.b.shape[0] == len(a0)
        self.abi = a0 @ self.b
    
    def iter_out(self, i):
        abi = self.abi
        if self.vec:
            abi = jnp.squeeze(abi, -1)
        return abi
    
    def iter(self, i, ai):
        """ i-th row of input @ b """
        self.abi = ai @ self.b
    
class SolveTriLowerColByFull(MatMulIterByFull):
    # x[0] /= a[0, 0]
    # for i in range(1, len(x)):
    #     x[i:] -= x[i - 1] * a[i:, i - 1]
    #     x[i] /= a[i, i]
        
    def init(self, n, a0):
        b = self.b
        del self.b
        assert a0.shape == (n,)
        assert b.shape[0] == n
        self.prevai = a0.at[0].set(0)
        self.ab = b.at[0, :].divide(a0[0])
    
    def iter(self, i, ai):
        ab = self.ab
        ab = ab - ab[i - 1, :] * self.prevai[:, None]
        self.ab = ab.at[i, :].divide(ai[i])
        self.prevai = ai.at[i].set(0)

class Rows(Producer):
    
    def __init__(self, x):
        self.x = x
    
    inputs = ()
    
    def init(self, n):
        pass
        
    def iter_out(self, i):
        return self.x[i, ...]
    
    def iter(self, i):
        pass

class MatMulColByRow(Consumer):
    
    def __init__(self, inputa, inputb):
        self.inputs = (inputa, inputb)
    
    inputs = None
    
    def init(self, n, a0, b0):
        assert a0.ndim == 1 and b0.ndim <= 1
        self.vec = b0.ndim > 0
        if self.vec:
            self.ab = a0[:, None] * b0[None, :]
        else:
            self.ab = a0 * b0
    
    def iter(self, i, ai, bi):
        if self.vec:
            self.ab = self.ab + ai[:, None] * bi[None, :]
        else:
            self.ab = self.ab + ai * bi
    
    def finalize(self):
        return self.ab

class SumLogDiag(Consumer, SingleInput):
    """input = operation producing the rows/columns of a square matrix"""
    
    def init(self, n, m0):
        assert m0.shape == (n,)
        self.sld = jnp.log(m0[0])
    
    def iter(self, i, mi):
        self.sld = self.sld + jnp.log(mi[i])
    
    def finalize(self):
        """sum(log(diag(m)))"""
        return self.sld



from scipy import linalg
import numpy as np
from matplotlib import pyplot as plt
import tqdm

# checks a symmetric nxn toeplitz matrix can be embedded in a circulant
# (2n-2)x(2n-2) matrix (its 2n-1 for non symmetric)

t = np.arange(6)[::-1]

c = np.concatenate([t, t[1:-1][::-1]])
# question: if t is positive definite, is c positive definite too? I think so
# because it's like we are stitching together repetitions of the process, but
# I'm not sure

m = linalg.toeplitz(c)
m2 = linalg.circulant(c)

np.testing.assert_equal(m2, m)

# check the fourier diagonalization of the circulant matrix (to be sure of
# the correct normalization)

d = np.diag(np.fft.fft(c, norm='backward'))
m3 = np.fft.fft(np.fft.fft(d, axis=0, norm='ortho').conj(), axis=1, norm='ortho')

np.testing.assert_allclose(m3, m, rtol=1e-14, atol=1e-14)

# try using rfft
# => doesn't work
#
# d = np.diag(np.fft.rfft(c, norm='backward'))
# m4 = np.fft.irfft(np.fft.irfft(d, axis=0, norm='ortho'), axis=1, norm='ortho')
#
# np.testing.assert_allclose(m4, m)

nmc = 10000
z = np.random.randn(len(c), nmc)
f = np.fft.fft(c, norm='backward')
y = np.fft.fft(np.sqrt(f)[:, None] * z, axis=0, norm='ortho')
x = y[:len(t), :]
x = x.real + x.imag # <-----
v = linalg.toeplitz(t)
v2 = np.cov(x)

a = np.linalg.norm(v2 - v, 2)
r = a / np.linalg.norm(v)
print(f'abs = {a:.2g}, relative = {r:.2g}')

# generate a random pos def toeplitz matrix
def randt_dumb(n):
    N = 2 * n
    m = np.random.randn(N, N)
    a = m @ m.T
    c = np.zeros_like(a)
    for i in range(N):
        # there is a faster way to generate only the first row
        c += np.roll(a, (i, i), (0, 1))
    t = c[:n, :n]
    return t[0]

def randt(n):
    N = 2 * n - 1
    x = np.random.randn(N)
    a = np.fft.ifft(np.abs(np.fft.fft(x)))
    assert np.allclose(a.imag, 0, atol=1e-15)
    t = a.real[:n]
    return t

def embed(t):
    c = np.concatenate([t, t[1:][::-1]])
    return c

eig = []
for i in tqdm.tqdm(range(100000)):
    t = randt(20)
    assert np.all(linalg.eigvalsh(linalg.toeplitz(t)) >= 0)
    c = embed(t)
    m = linalg.circulant(c)
    e = linalg.eigvalsh(m)
    assert np.all(e >= 0)
    eig.append(e)

fig, ax = plt.subplots(num='embed', clear=True)

e = np.concatenate(eig)
ax.hist(e, bins='auto', histtype='step')

fig.show()



# lsqfitgp/_kernels/_arma.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import jax
from jax import numpy as jnp
from jax import lax
import numpy

from .. import _linalg
from .. import _patch_jax
from .._Kernel import stationarykernel

# use positive delta because negative indices wrap around
@stationarykernel(derivable=False, maxdim=1, input='hard')
def MA(delta, w=None, norm=False):
    """
    Discrete moving average kernel.
    
    .. math::
        k(\\Delta) = \\sum_{k=|\\Delta|}^{n-1} w_k w_{k-|\\Delta|},
        \\quad \\mathbf w = (w_0, \\ldots, w_{n-1}).
    
    The inputs must be integers. It is the autocovariance function of a moving
    average with weights :math:`\\mathbf w` applied to white noise:
    
    .. math::
        k(i, j) &= \\operatorname{Cov}[y_i, y_j], \\\\
        y_i &= \\sum_{k=0}^{n-1} w_k \\epsilon_{i-k}, \\\\
        \\operatorname{Cov}[\\epsilon_i,\\epsilon_j] &= \\delta_{ij}.
    
    If `norm` is True, the variance is normalized to 1, which amounts to
    normalizing :math:`\\mathbf w` to unit length.
    
    """
    
    # TODO reference? must find some standard book with a treatment which is
    # not too formal yet writes clearly about the covariance function
    
    # TODO nd version with w.ndim == n, it's a nd convolution. use
    # jax.scipy.signal.correlate.
        
    w = jnp.asarray(w)
    assert w.ndim == 1
    if len(w):
        cov = jnp.convolve(w, w[::-1])
        if norm:
            cov /= cov[len(w) - 1]
        return cov.at[delta + len(w) - 1].get(mode='fill', fill_value=0)
    else:
        return jnp.zeros(delta.shape)

@stationarykernel(derivable=False, maxdim=1, input='hard')
def _ARBase(delta, phi=None, gamma=None, maxlag=None, slnr=None, lnc=None, norm=False):
    """
    Discrete autoregressive kernel.
        
    You have to specify one and only one of the sets of parameters
    `phi+maxlag`, `gamma+maxlag`, `slnr+lnc`.

    Parameters
    ----------
    phi : (p,) real
        The autoregressive coefficients at lag 1...p.
    gamma : (p + 1,) real
        The autocovariance at lag 0...p.
    maxlag : int
        The maximum lag that the kernel will be evaluated on. If the actual
        inputs produce higher lags, the missing values are filled with ``nan``.
    slnr : (nr,) real
        The real roots of the characteristic polynomial, expressed in the
        following way: ``sign(slnr)`` is the sign of the root, and
        ``abs(snlr)`` is the natural logarithm of the absolute value.
    lnc : (nc,) complex
        The natural logarithm of the complex roots of the characteristic
        polynomial (:math:`\\log z = \\log|z| + i\\arg z`), where each root
        also stands for its paired conjugate.
    
        In `slnr` and `lnc`, the multiplicity of a root is expressed by
        repeating the root in the array (not necessarily next to each other).
        Only exact repetition counts; very close yet distinct roots are treated
        as separate and lead to numerical instability, in particular complex
        roots very close to the real line. An exactly real complex root behaves
        like a pair of identical real roots. Two complex roots also count as
        equal if conjugate, and the argument is standardized to :math:`[0,
        2\\pi)`.
    norm : bool
        If True, normalize the autocovariance to be 1 at lag 0. If False
        (default), normalize such that the variance of the generating noise is
        1, or use the user-provided normalization if `gamma` is specified.
    
    Notes
    -----
    This is the covariance function of a stationary autoregressive process,
    which is defined recursively as
    
    .. math::
        y_i = \\sum_{k=1}^p \\phi_k y_{i-k} + \\epsilon_i,
    
    where :math:`\\epsilon_i` is white noise, i.e.,
    :math:`\\operatorname{Cov}[\\epsilon_i, \\epsilon_j] = \\delta_{ij}`. The
    length :math:`p` of the vector of coefficients :math:`\\boldsymbol\\phi`
    is the `order` of the process.
    
    The covariance function can be expressed in two ways. First as the same
    recursion defining the process:
    
    .. math::
        \\gamma_m = \\sum_{k=1}^p \\phi_k \\gamma_{m-k} + \\delta_{m0},
    
    where :math:`\\gamma_m \\equiv \\operatorname{Cov}[y_i, y_{i+m}]`. This is
    called `Yule-Walker equation`. Second, as a linear combination of mixed
    power-exponentials:
    
    .. math::
        \\gamma_m = \\sum_{j=1}^n
                    \\sum_{l=1}^{\\mu_j}
                    a_{jl} |m|^{l-1} x_j^{-|m|},
    
    where :math:`x_j` and :math:`\\mu_j` are the (complex) roots and
    corresponding multiplicities of the `characteristic polynomial`
    
    .. math::
        P(x) = 1 - \\sum_{k=1}^p \\phi_k x^k,
    
    and the :math:`a_{jl}` are uniquely determined complex coefficients. The
    :math:`\\boldsymbol\\phi` vector is valid iff :math:`|x_j|>1, \\forall j`.
    
    There are three alternative parametrization for this kernel.
    
    If you specify `phi`, the first terms of the covariance are computed
    solving the Yule-Walker equation, and then evolved up to `maxlag`. It
    is necessary to specify `maxlag` instead of letting the code figure it out
    from the actual inputs for technical reasons.
    
    Likewise, if you specify `gamma`, the coefficients are obtained with
    Yule-Walker and then used to evolve the covariance. The only difference is
    that the normalization can be different: starting from `phi`, the variance
    of the generating noise :math:`\\epsilon` is fixed to 1, while giving
    `gamma` directly implies an arbitrary value.
    
    Instead, if you specify the roots with `slnr` and `lnc`, the coefficients
    are obtained from the polynomial defined in terms of the roots, and then
    the amplitudes :math:`a_{jl}` are computed by solving a linear system with
    the covariance (from YW) as RHS. Finally, the full covariance function is
    evaluated with the analytical expression.
    
    The reasons for using the logarithm are that 1) in practice the roots are
    tipically close to 1, so the logarithm is numerically more accurate, and 2)
    the logarithm is readily interpretable as the inverse of the correlation
    length.
    
    """
    cond = (
        (phi is not None and maxlag is not None and gamma is None and slnr is None and lnc is None) or
        (phi is None and maxlag is not None and gamma is not None and slnr is None and lnc is None) or
        (phi is None and maxlag is None and gamma is None and slnr is not None and lnc is not None)
    )
    if not cond:
        raise ValueError('invalid set of specified parameters')
    
    # TODO maybe I could allow redundantly specifying gamma and phi, e.g., for
    # numerical accuracy reasons if they are determined from an analytical
    # expression.
    
    if phi is None and gamma is None:
        return _ar_with_roots(delta, slnr, lnc, norm)
    else:
        return _ar_with_phigamma(delta, phi, gamma, maxlag, norm)

def _ar_with_phigamma(delta, phi, gamma, maxlag, norm):
    if phi is None:
        phi = AR.phi_from_gamma(gamma)
    if gamma is None:
        gamma = AR.gamma_from_phi(phi)
    if norm:
        gamma = gamma / gamma[0]
    acf = AR.extend_gamma(gamma, phi, maxlag + 1 - len(gamma))
    return acf.at[delta].get(mode='fill', fill_value=jnp.nan)

def _yule_walker(gamma):
    """
    gamma = autocovariance at lag 0...p
    output: autoregressive coefficients at lag 1...p
    """
    gamma = jnp.asarray(gamma)
    assert gamma.ndim == 1
    t = gamma[:-1]
    b = gamma[1:]
    if t.size:
        return _linalg._toeplitz.solve(t, b)
    else:
        return jnp.empty(0)

def _yule_walker_inv_mat(phi):
    phi = jnp.asarray(phi)
    assert phi.ndim == 1
    p = len(phi)
    m = jnp.arange(p + 1)[:, None] # rows
    n = m.T # columns
    phi = jnp.pad(phi, (1, 1))
    kp = jnp.clip(m + n, 0, p + 1)
    km = jnp.clip(m - n, 0, p + 1)
    return jnp.eye(p + 1) - (phi[kp] + phi[km]) / jnp.where(n, 1, 2)
    
def _yule_walker_inv(phi):
    """
    phi = autoregressive coefficients at lag 1...p
    output: autocovariance at lag 0...p, assuming driving noise has sdev 1
    """
    a = _yule_walker_inv_mat(phi)
    b = jnp.zeros(len(a)).at[0].set(1)
    # gamma = _pseudo_solve(a, b)
    gamma = jnp.linalg.solve(a, b)
    return gamma

def _ar_evolve(phi, start, noise):
    """
    phi = autoregressive coefficients at lag 1...p
    start = first p values of the process (increasing time)
    noise = n noise values added at each step
    output: n new process values
    """
    phi = jnp.asarray(phi)
    start = jnp.asarray(start)
    noise = jnp.asarray(noise)
    assert phi.ndim == 1 and phi.shape == start.shape and noise.ndim == 1
    return _ar_evolve_jit(phi, start, noise)

@jax.jit
def _ar_evolve_jit(phi, start, noise):
    
    def f(carry, eps):
        vals, cc, roll = carry
        phi = lax.dynamic_slice(cc, [vals.size - roll], [vals.size])
        nextval = phi @ vals + eps
        if vals.size:
            vals = vals.at[roll].set(nextval)
        # maybe for some weird reason like alignment, actual rolling would
        # be faster. whatever
        roll = (roll + 1) % vals.size
        return (vals, cc, roll), nextval
    
    cc = jnp.concatenate([phi, phi])[::-1]
    _, ev = lax.scan(f, (start, cc, 0), noise, unroll=16)
    return ev

def _ar_with_roots(delta, slnr, lnc, norm):
    phi = AR.phi_from_roots(slnr, lnc) # <---- weak
    gamma = AR.gamma_from_phi(phi) # <---- point
    if norm:
        gamma /= gamma[0]
    ampl = AR.ampl_from_roots(slnr, lnc, gamma)
    acf = AR.cov_from_ampl(slnr, lnc, ampl, delta)
    return acf

    # TODO Currently gamma is not even pos def for high multiplicity/roots close
    # to 1. Raw patch: the badness condition is gamma[0] < 0 or any(abs(gamma) >
    # gamma[0]) or gamma inf/nan. Take the smallest log|root| and assume it
    # alone determines gamma. This is best implemented as an option in
    # _gamma_from_ampl_matmul.
    
    # Is numerical integration of the spectrum a feasible way to get the
    # covariance? The roots correspond to peaks, and they get very high as the
    # roots get close to 1. But I know where the peaks are in advance => nope
    # because the e^iwx oscillates arbitrarily fast. However maybe I can compute
    # the first p terms, which solves my current problem with gamma. I guess I
    # just have to use a multiple of p of quadrature points. The spectrum
    # oscillates too but only up to mode p. The total calculation cost is then
    # O(p^2), better than current O(p^3). See Hamilton (1994, p. 155).
    
    # Other solution (Hamilton p. 319): the covariance should be equal to the
    # impulse response, so I can get gamma from phi by an evolution starting
    # from zeros. => Nope, it's equal only for AR(1).
    
    # condition for phi: in the region phi >= 0, it must be sum(phi) <= 1
    # (Hamilton p. 659).
    
    # p = phi.size
    # yw = _yule_walker_inv_mat(phi)
    # b = jnp.zeros(p + 1).at[0].set(1)
    # ampl = jnp.linalg.solve(yw @ mat, b)
    # lag = delta if delta.ndim else delta[None]
    # acf = _gamma_from_ampl_matmul(slnr, lnc, lag, ampl)
    # if norm:
    #     acf0 = _gamma_from_ampl_matmul(slnr, lnc, jnp.array([0]), ampl)
    #     acf /= acf0
    # return acf if delta.ndim else acf.squeeze(0)

def _pseudo_solve(a, b):
    # this is less accurate than jnp.linalg.solve
    u, s, vh = jnp.linalg.svd(a)
    eps = jnp.finfo(a.dtype).eps
    s0 = s[0] if s.size else 0
    invs = jnp.where(s < s0 * eps * len(a), 0, 1 / s)
    return jnp.einsum('ij,j,jk,k', vh.conj().T, invs, u.conj().T, b)

@jax.jit
def _gamma_from_ampl_matmul(slnr, lnc, lag, ampl, lagnorm=None):

    vec = ampl.ndim == 1
    if vec:
        ampl = ampl[:, None]
    p = slnr.size + 2 * lnc.size
    assert ampl.shape[-2] == p + 1
    if lagnorm is None:
        lagnorm = p
    
    def logcol(root, lag, llag, repeat):
        return -root * lag + jnp.where(repeat, repeat * llag, 0)
    
    def lognorm(root, repeat, lagnorm):
        maxnorm = jnp.where(repeat, repeat * (-1 + jnp.log(repeat / root)), 0)
        defnorm = logcol(root, lagnorm, jnp.log(lagnorm), repeat)
        maxloc = repeat / root
        return jnp.where(maxloc <= lagnorm, maxnorm, defnorm)
    
    # roots at infinity
    # TODO remove this because it's degenerate with large roots, handle the
    # p=0 case outside of this function
    col = jnp.where(lag, 0, 1)
    out = col[..., :, None] * ampl[..., 0, :]
    
    # real roots
    llag = jnp.log(lag)
    val = (jnp.nan, 0, out, slnr, lag, llag, lagnorm)
    def loop(i, val):
        prevroot, repeat, out, slnr, lag, llag, lagnorm = val
        root = slnr[i]
        repeat = jnp.where(root == prevroot, repeat + 1, 0)
        prevroot = root
        sign = jnp.sign(root) ** lag
        aroot = jnp.abs(root)
        lcol = logcol(aroot, lag, llag, repeat)
        norm = lognorm(aroot, repeat, lagnorm)
        col = sign * jnp.exp(lcol - norm)
        out += col[..., :, None] * ampl[..., 1 + i, :]
        return prevroot, repeat, out, slnr, lag, llag, lagnorm
    if slnr.size:
        _, _, out, _, _, _, _ = lax.fori_loop(0, slnr.size, loop, val)
    
    # complex roots
    val = (jnp.nan, 0, out, lnc, lag, llag, lagnorm)
    def loop(i, val):
        prevroot, repeat, out, lnc, lag, llag, lagnorm = val
        root = lnc[i]
        repeat = jnp.where(root == prevroot, repeat + 1, 0)
        prevroot = root
        lcol = logcol(root, lag, llag, repeat)
        norm = lognorm(root.real, repeat, lagnorm)
        col = jnp.exp(lcol - norm)
        idx = 1 + slnr.size + 2 * i
        out += col.real[..., :, None] * ampl[..., idx, :]
        
        # real complex root = a pair of identical real roots
        repeat = jnp.where(root.imag, repeat, repeat + 1)
        col1 = jnp.where(root.imag, -col.imag, col.real * lag)
        out += col1[..., :, None] * ampl[..., idx + 1, :]
        
        return prevroot, repeat, out, lnc, lag, llag, lagnorm
    if lnc.size:
        _, _, out, _, _, _, _ = lax.fori_loop(0, lnc.size, loop, val)
    
    if vec:
        out = out.squeeze(-1)
    
    return out

class AR(_ARBase):
    
    __doc__ = _ARBase.__doc__
    
    @classmethod
    def phi_from_gamma(cls, gamma):
        """
        Determine the autoregressive coefficients from the covariance.
        
        Parameters
        ----------
        gamma : (p + 1,) array
            The autocovariance at lag 0...p.
        
        Returns
        -------
        phi : (p,) array
            The autoregressive coefficients at lag 1...p.
        """
        gamma = cls._process_gamma(gamma)
        return _yule_walker(gamma)
    
    @classmethod
    def gamma_from_phi(cls, phi):
        """
        Determine the covariance from the autoregressive coefficients.

        Parameters
        ----------
        phi : (p,) array
            The autoregressive coefficients at lag 1...p.
        
        Returns
        -------
        gamma : (p + 1,) array
            The autocovariance at lag 0...p. The normalization is
            with noise variance 1.
        
        Notes
        -----
        The result is wildly inaccurate for roots with high multiplicity and/or
        close to 1.
        """
        phi = cls._process_phi(phi)
        return _yule_walker_inv(phi)
        
        # TODO fails (nan) for very small roots. In that case the answer is that
        # gamma is a constant vector. However I can't get the constant out of
        # a degenerate phi, I need the roots, and I don't know the formula.
    
    @classmethod
    def extend_gamma(cls, gamma, phi, n):
        """
        Extends values of the covariance function to higher lags.
        
        Parameters
        ----------
        gamma : (m,) array
            The autocovariance at lag q-m+1...q, with q >= 0 and m >= p + 1.
        phi : (p,) array
            The autoregressive coefficients at lag 1...p.
        n : int
            The number of new values to generate.
        
        Returns
        -------
        ext : (m + n,) array
            The autocovariance at lag q-m+1...q+n.
        """
        gamma = cls._process_gamma(gamma)
        phi = cls._process_phi(phi)
        assert gamma.size > phi.size
        ext = _ar_evolve(phi, gamma[len(gamma) - len(phi):], jnp.broadcast_to(0., (n,)))
        return jnp.concatenate([gamma, ext])
    
    @classmethod
    def phi_from_roots(cls, slnr, lnc):
        """
        Determine the autoregressive coefficients from the roots of the
        characteristic polynomial.
        
        Parameters
        ----------
        slnr : (nr,) real
            The real roots of the characteristic polynomial, expressed in the
            following way: ``sign(slnr)`` is the sign of the root, and
            ``abs(snlr)`` is the natural logarithm of the absolute value.
        lnc : (nc,) complex
            The natural logarithm of the complex roots of the characteristic
            polynomial (:math:`\\log z = \\log|z| + i\\arg z`), where each root
            also stands for its paired conjugate.
        
        Returns
        -------
        phi : (p,) real
            The autoregressive coefficients at lag 1...p, with p = nr + 2 nc.
        """
        slnr, lnc = cls._process_roots(slnr, lnc)
        r = jnp.copysign(jnp.exp(-jnp.abs(slnr)), slnr) # works with +/-0
        c = jnp.exp(-lnc)
        
        # minus sign in the exponentials to do 1/z, the poly output is already
        # reversed
                
        roots = jnp.concatenate([r, c, c.conj()]).sort() # <-- polyroots sorts
        coef = jnp.atleast_1d(jnp.poly(roots))
        
        # TODO the implementation of jnp.poly (and np.poly) is inferior to the
        # one of np.polynomial.polynomial.polyfromroots, which cares about
        # numerical accuracy and would reduce compilation time if ported to jax
        # (current one is O(p), that would be O(log p)).
        
        if coef.size:
            with _patch_jax.skipifabstract():
                numpy.testing.assert_equal(coef[0].item(), 1)
                numpy.testing.assert_allclose(jnp.imag(coef), 0, rtol=0, atol=1e-4)
        return -coef.real[1:]
        
        # TODO possibly not accurate for large p. Do a test with an
        # implementation of poly which uses integer roots and non-fft convolve
        # (maybe add it as an option to my to-be-written implementation of poly)

    @classmethod
    def ampl_from_roots(cls, slnr, lnc, gamma):
        # TODO docs
        slnr, lnc = cls._process_roots(slnr, lnc)
        gamma = cls._process_gamma(gamma)
        assert gamma.size == 1 + slnr.size + 2 * lnc.size
        lag = jnp.arange(gamma.size)
        mat = _gamma_from_ampl_matmul(slnr, lnc, lag, jnp.eye(gamma.size))
        # return jnp.linalg.solve(mat, gamma)
        return _pseudo_solve(mat, gamma)
        
        # TODO I'm using pseudo-solve only because of large roots degeneracy
        # in _gamma_from_ampl_matmul, remove it after solving that
        
        # TODO maybe I can increase the precision of the solve with some
        # ordering of the columns of mat, I guess (reversed) global sort of the
        # roots
    
    @classmethod
    def cov_from_ampl(cls, slnr, lnc, ampl, lag):
        # TODO docs
        slnr, lnc = cls._process_roots(slnr, lnc)
        ampl = cls._process_ampl(ampl)
        assert ampl.size == 1 + slnr.size + 2 * lnc.size
        lag = cls._process_lag(lag)
        scalar = lag.ndim == 0
        if scalar:
            lag = lag[None]
        acf = _gamma_from_ampl_matmul(slnr, lnc, lag, ampl)
        return acf.squeeze(0) if scalar else acf
        
    @classmethod
    def inverse_roots_from_phi(cls, phi):
        phi = cls._process_phi(phi)
        poly = jnp.concatenate([jnp.ones(1), -phi])
        return jnp.roots(poly, strip_zeros=False)
    
    # TODO methods:
    # - gamma_from_roots which uses quadrature fourier transf of spectrum
    
    @staticmethod
    def _process_roots(slnr, lnc):
        slnr = jnp.asarray(slnr, float).sort()
        lnc = jnp.asarray(lnc, complex)
        assert slnr.ndim == lnc.ndim == 1
        imag = jnp.abs(lnc.imag) % (2 * jnp.pi)
        imag = jnp.where(imag > jnp.pi, 2 * jnp.pi - imag, imag)
        lnc = lnc.real + 1j * imag
        lnc = lnc.sort()
        return slnr, lnc
    
    @staticmethod
    def _process_gamma(gamma):
        gamma = jnp.asarray(gamma, float)
        assert gamma.ndim == 1 and gamma.size >= 1
        return gamma
    
    @staticmethod
    def _process_phi(phi):
        phi = jnp.asarray(phi, float)
        assert phi.ndim == 1
        return phi
    
    @staticmethod
    def _process_ampl(ampl):
        ampl = jnp.asarray(ampl, float)
        assert ampl.ndim == 1 and ampl.size >= 1
        return ampl
    
    @staticmethod
    def _process_lag(lag):
        lag = jnp.asarray(lag)
        assert jnp.issubdtype(lag, jnp.integer)
        return lag.astype(int)

